from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.execution_status import ExecutionStatus
from ..models.rule_operator import RuleOperator
from ..types import UNSET, Unset

T = TypeVar("T", bound="RuleResult")


@_attrs_define
class RuleResult:
    """
    Attributes
    ----------
        metric (str): Name of the metric.
        operator (RuleOperator):
        target_value (Union[None, float, int, list[Any], str]): Value to compare with for this metric (right hand side).
        status (Union[Unset, ExecutionStatus]): Status of the execution.
        value (Union[Any, None, Unset]): Result of the metric computation.
        execution_time (Union[None, Unset, float]): Execution time for the rule in seconds.
    """

    metric: str
    operator: RuleOperator
    target_value: Union[None, float, int, list[Any], str]
    status: Union[Unset, ExecutionStatus] = UNSET
    value: Union[Any, None, Unset] = UNSET
    execution_time: Union[None, Unset, float] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        metric = self.metric

        operator = self.operator.value

        target_value: Union[None, float, int, list[Any], str]
        target_value = self.target_value if isinstance(self.target_value, list) else self.target_value

        status: Union[Unset, str] = UNSET
        if not isinstance(self.status, Unset):
            status = self.status.value

        value: Union[Any, None, Unset]
        value = UNSET if isinstance(self.value, Unset) else self.value

        execution_time: Union[None, Unset, float]
        execution_time = UNSET if isinstance(self.execution_time, Unset) else self.execution_time

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"metric": metric, "operator": operator, "target_value": target_value})
        if status is not UNSET:
            field_dict["status"] = status
        if value is not UNSET:
            field_dict["value"] = value
        if execution_time is not UNSET:
            field_dict["execution_time"] = execution_time

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        metric = d.pop("metric")

        operator = RuleOperator(d.pop("operator"))

        def _parse_target_value(data: object) -> Union[None, float, int, list[Any], str]:
            if data is None:
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                return cast(list[Any], data)

            except:  # noqa: E722
                pass
            return cast(Union[None, float, int, list[Any], str], data)

        target_value = _parse_target_value(d.pop("target_value"))

        _status = d.pop("status", UNSET)
        status: Union[Unset, ExecutionStatus]
        status = UNSET if isinstance(_status, Unset) else ExecutionStatus(_status)

        def _parse_value(data: object) -> Union[Any, None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[Any, None, Unset], data)

        value = _parse_value(d.pop("value", UNSET))

        def _parse_execution_time(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        execution_time = _parse_execution_time(d.pop("execution_time", UNSET))

        rule_result = cls(
            metric=metric,
            operator=operator,
            target_value=target_value,
            status=status,
            value=value,
            execution_time=execution_time,
        )

        rule_result.additional_properties = d
        return rule_result

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Experiment result wrapper for easy status access and monitoring."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from galileo.resources.types import Unset

if TYPE_CHECKING:
    from galileo.resources.models import ExperimentResponse

logger = logging.getLogger(__name__)


class ExperimentPhaseInfo:
    """
    Wrapper for experiment phase status information.

    Provides easy access to phase-level progress and status information for an experiment.

    Attributes
    ----------
        progress_percent (float): Progress percentage (0.0 to 100.0).
        is_complete (bool): Whether this phase is complete.
        is_failed (bool): Whether this phase failed.
        is_in_progress (bool): Whether this phase is in progress.

    Examples
    --------
        # Access phase information from experiment status
        experiment = Experiment.get(name="ml-eval", project_name="My Project")
        status = experiment.get_status()

        log_gen = status.log_generation
        if log_gen.is_complete:
            print("Log generation phase completed!")
        elif log_gen.is_in_progress:
            print(f"Log generation at {log_gen.progress_percent:.1f}%")

        # Check phase status
        print(f"Phase status: {log_gen}")  # Human-readable string
    """

    def __init__(self, phase_status: Any):
        """
        Initialize from ExperimentPhaseStatus object.

        Args:
            phase_status: The phase status object from the API response.
        """
        self.progress_percent = getattr(phase_status, "progress_percent", 0.0)
        if isinstance(self.progress_percent, Unset):
            self.progress_percent = 0.0
        # Convert from API's 0.0-1.0 range to 0.0-100.0 percentage
        self.progress_percent = self.progress_percent * 100.0

    @property
    def is_complete(self) -> bool:
        """Whether this phase is complete (progress >= 100%)."""
        return self.progress_percent >= 100.0

    @property
    def is_failed(self) -> bool:
        """Whether this phase failed. Currently always returns False."""
        # TODO: Check for actual failure status when available in API
        return False

    @property
    def is_in_progress(self) -> bool:
        """Whether this phase is in progress (0% < progress < 100%)."""
        return 0.0 < self.progress_percent < 100.0

    @property
    def is_pending(self) -> bool:
        """Whether this phase hasn't started yet (progress = 0%)."""
        return self.progress_percent == 0.0

    def to_dict(self) -> dict[str, Any]:
        """
        Convert the phase info to a dictionary.

        Returns
        -------
            A dictionary containing phase status information.

        Examples
        --------
            phase_dict = status.log_generation.to_dict()
            print(phase_dict["progress_percent"])
            print(phase_dict["is_complete"])
        """
        return {
            "progress_percent": self.progress_percent,
            "is_complete": self.is_complete,
            "is_failed": self.is_failed,
            "is_in_progress": self.is_in_progress,
            "is_pending": self.is_pending,
        }

    def __repr__(self) -> str:
        """Detailed string representation."""
        return (
            f"ExperimentPhaseInfo(progress_percent={self.progress_percent:.1f}, "
            f"is_complete={self.is_complete}, is_in_progress={self.is_in_progress})"
        )

    def __str__(self) -> str:
        """Human-readable string representation."""
        if self.is_complete:
            return "Complete (100%)"
        if self.is_in_progress:
            return f"In Progress ({self.progress_percent:.1f}%)"
        return "Not Started (0%)"


class ExperimentStatusInfo:
    """
    Wrapper for experiment status information.

    Provides human-readable access to experiment execution status, including progress
    tracking and phase-level information.

    Attributes
    ----------
        log_generation (ExperimentPhaseInfo): Status of log generation phase.
        overall_progress (float): Overall progress percentage (0.0 to 100.0).
        is_complete (bool): Whether the experiment is complete.
        is_in_progress (bool): Whether the experiment is currently running.
        is_pending (bool): Whether the experiment hasn't started yet.

    Examples
    --------
        # Get status information
        experiment = Experiment.get(name="ml-eval", project_name="My Project")
        status = experiment.get_status()

        # Human-readable output
        print(status)  # "Experiment Running (45.2%)"
        print(f"Log Generation: {status.log_generation}")  # "In Progress (45.2%)"

        # Check status
        if status.is_complete:
            print("Experiment completed!")
        elif status.is_in_progress:
            print(f"Progress: {status.overall_progress:.1f}%")
        elif status.is_pending:
            print("Experiment hasn't started yet")

        # Convert to dictionary
        status_dict = status.to_dict()
        print(status_dict["overall_progress"])
    """

    def __init__(self, experiment_response: ExperimentResponse):
        """
        Initialize from ExperimentResponse object.

        Args:
            experiment_response: The experiment response from the API.
        """
        self._response = experiment_response

        # Extract status info with proper handling of Unset types
        status_obj = getattr(experiment_response, "status", None)
        if status_obj is not None and not isinstance(status_obj, Unset):
            log_gen = getattr(status_obj, "log_generation", None)
            if log_gen is not None and not isinstance(log_gen, Unset):
                self.log_generation = ExperimentPhaseInfo(log_gen)
            else:
                self.log_generation = ExperimentPhaseInfo(type("obj", (), {"progress_percent": 0.0})())
        else:
            self.log_generation = ExperimentPhaseInfo(type("obj", (), {"progress_percent": 0.0})())

    @classmethod
    def from_result(cls, result: dict[str, Any]) -> ExperimentStatusInfo:
        """
        Create from experiment.run() result dictionary.

        Args:
            result: Dictionary returned from Experiment.run().

        Returns
        -------
            ExperimentStatusInfo: Status information extracted from the result.

        Examples
        --------
            result = experiment.run()
            status = ExperimentStatusInfo.from_result(result)
            print(f"Status: {status}")
        """
        return cls(result["experiment"])

    @property
    def overall_progress(self) -> float:
        """
        Overall progress percentage across all phases.

        Currently uses log_generation progress. In the future, this may
        average multiple phase progress values.

        Returns
        -------
            float: Progress percentage from 0.0 to 100.0.
        """
        # For now just use log_generation progress
        # Could average multiple phases in the future
        return self.log_generation.progress_percent

    @property
    def is_complete(self) -> bool:
        """Whether the experiment is complete (all phases at 100%)."""
        return self.log_generation.is_complete

    @property
    def is_in_progress(self) -> bool:
        """Whether the experiment is currently running (0% < progress < 100%)."""
        return self.log_generation.is_in_progress

    @property
    def is_pending(self) -> bool:
        """Whether the experiment hasn't started yet (progress = 0%)."""
        return self.overall_progress == 0.0

    @property
    def is_failed(self) -> bool:
        """Whether the experiment has failed. Currently always returns False."""
        # TODO: Check for actual failure status when available in API
        return self.log_generation.is_failed

    def to_dict(self) -> dict[str, Any]:
        """
        Convert the status info to a dictionary.

        Returns
        -------
            A dictionary containing status information.

        Examples
        --------
            status = experiment.get_status()
            status_dict = status.to_dict()
            print(status_dict["overall_progress"])
            print(status_dict["log_generation"]["is_complete"])
        """
        return {
            "overall_progress": self.overall_progress,
            "is_complete": self.is_complete,
            "is_in_progress": self.is_in_progress,
            "is_pending": self.is_pending,
            "is_failed": self.is_failed,
            "log_generation": self.log_generation.to_dict(),
        }

    def __repr__(self) -> str:
        """Detailed string representation."""
        return (
            f"ExperimentStatusInfo(\n"
            f"  log_generation={self.log_generation!r},\n"
            f"  overall_progress={self.overall_progress:.1f}%,\n"
            f"  is_complete={self.is_complete},\n"
            f"  is_in_progress={self.is_in_progress}\n"
            f")"
        )

    def __str__(self) -> str:
        """Human-readable string representation."""
        if self.is_complete:
            return "Experiment Complete"
        if self.is_in_progress:
            return f"Experiment Running ({self.overall_progress:.1f}%)"
        return "Experiment Pending"


class ExperimentRunResult:
    """
    Wrapper for experiment.run() result with human-readable access.

    Provides easy access to experiment run information including the link,
    status, and underlying experiment response. This wrapper makes it simple
    to work with experiment run results by exposing commonly needed information
    through intuitive properties and methods.

    Attributes
    ----------
        link (str): URL to view experiment results in Galileo console.
        message (str): Status message about the experiment.
        status (ExperimentStatusInfo): Detailed status information.
        experiment_id (str): The experiment ID.
        project_id (str): The project ID.
        experiment_name (str): The experiment name.

    Examples
    --------
        # Run an experiment and access the result
        experiment = Experiment(
            name="ml-evaluation",
            dataset_name="ml-dataset",
            prompt_name="ml-prompt",
            project_name="My AI Project"
        ).create()

        result = experiment.run()

        # Access basic information
        print(result)  # Human-readable summary
        print(f"View results: {result.link}")
        print(f"Experiment ID: {result.experiment_id}")

        # Check status
        if result.status.is_in_progress:
            print(f"Progress: {result.status.overall_progress:.1f}%")
        elif result.status.is_complete:
            print("Experiment completed!")

        # Get dataset and prompt information
        if result.dataset_info:
            print(f"Dataset: {result.dataset_info['name']}")
        if result.prompt_info:
            print(f"Prompt: {result.prompt_info['name']}")

        # Convert to dictionary
        result_dict = result.to_dict()
        print(result_dict["link"])
    """

    def __init__(self, result: dict[str, Any]):
        """
        Initialize from experiment.run() result.

        Args:
            result: Dictionary returned from Experiment.run().

        Raises
        ------
            KeyError: If required keys are missing from result dictionary.
            TypeError: If result is not a dictionary.
        """
        if not isinstance(result, dict):
            raise TypeError(f"Expected dict, got {type(result).__name__}")

        if "experiment" not in result:
            raise KeyError("Result dictionary missing required key 'experiment'")

        self._result = result
        self._experiment_response: ExperimentResponse = result["experiment"]
        self.link = result.get("link", "")
        self.message = result.get("message", "")
        self.status = ExperimentStatusInfo(self._experiment_response)
        self.experiment_id = self._experiment_response.id
        self.project_id = self._experiment_response.project_id
        self.experiment_name = getattr(self._experiment_response, "name", "")

    @property
    def experiment(self) -> ExperimentResponse:
        """
        Get the underlying ExperimentResponse object.

        Returns
        -------
            ExperimentResponse: The raw API response object.
        """
        return self._experiment_response

    @property
    def dataset_info(self) -> dict[str, str | None] | None:
        """
        Get dataset information if available.

        Returns
        -------
            dict or None: Dictionary with dataset id, name, and version, or None if no dataset.

        Examples
        --------
            result = experiment.run()
            if result.dataset_info:
                print(f"Dataset: {result.dataset_info['name']}")
                print(f"Version: {result.dataset_info['version']}")
        """
        dataset = getattr(self._experiment_response, "dataset", None)
        if dataset is None or isinstance(dataset, Unset):
            return None

        return {
            "id": getattr(dataset, "dataset_id", None),
            "name": getattr(dataset, "name", None),
            "version": getattr(dataset, "version_index", None),
        }

    @property
    def prompt_info(self) -> dict[str, str | None] | None:
        """
        Get prompt information if available.

        Returns
        -------
            dict or None: Dictionary with prompt id and name, or None if no prompt.

        Examples
        --------
            result = experiment.run()
            if result.prompt_info:
                print(f"Prompt: {result.prompt_info['name']}")
                print(f"ID: {result.prompt_info['id']}")
        """
        prompt = getattr(self._experiment_response, "prompt", None)
        if prompt is None or isinstance(prompt, Unset):
            return None

        return {"id": getattr(prompt, "prompt_template_id", None), "name": getattr(prompt, "name", None)}

    def to_dict(self) -> dict[str, Any]:
        """
        Convert the result to a dictionary.

        Returns
        -------
            A dictionary containing all result information.

        Examples
        --------
            result = experiment.run()
            result_dict = result.to_dict()
            print(result_dict["link"])
            print(result_dict["status"]["overall_progress"])
        """
        return {
            "experiment_id": self.experiment_id,
            "experiment_name": self.experiment_name,
            "project_id": self.project_id,
            "link": self.link,
            "message": self.message,
            "status": self.status.to_dict(),
            "dataset_info": self.dataset_info,
            "prompt_info": self.prompt_info,
        }

    def __repr__(self) -> str:
        """Detailed string representation."""
        return (
            f"ExperimentRunResult(\n"
            f"  experiment_id='{self.experiment_id}',\n"
            f"  experiment_name='{self.experiment_name}',\n"
            f"  project_id='{self.project_id}',\n"
            f"  status={self.status!r},\n"
            f"  link='{self.link}',\n"
            f"  dataset_info={self.dataset_info},\n"
            f"  prompt_info={self.prompt_info}\n"
            f")"
        )

    def __str__(self) -> str:
        """Human-readable string representation."""
        lines = [
            "=" * 70,
            f"Experiment: {self.experiment_name}",
            "=" * 70,
            f"Status: {self.status}",
            f"Link: {self.link}",
            "",
        ]

        if self.dataset_info:
            dataset_name = self.dataset_info.get("name", "N/A")
            dataset_id = self.dataset_info.get("id", "N/A")
            lines.append(f"Dataset: {dataset_name} (id: {dataset_id})")

        if self.prompt_info:
            prompt_name = self.prompt_info.get("name", "N/A")
            prompt_id = self.prompt_info.get("id", "N/A")
            lines.append(f"Prompt: {prompt_name} (id: {prompt_id})")

        lines.append("=" * 70)

        return "\n".join(lines)



from enum import Enum


class OutputTypeEnum(str, Enum):
    BOOLEAN = "boolean"
    CATEGORICAL = "categorical"
    COUNT = "count"
    DISCRETE = "discrete"
    FREEFORM = "freeform"
    MULTILABEL = "multilabel"
    PERCENTAGE = "percentage"

    def __str__(self) -> str:
        return str(self.value)



from typing import Union

from galileo.resources.models import (
    LogRecordsBooleanFilter,
    LogRecordsDateFilter,
    LogRecordsIDFilter,
    LogRecordsNumberFilter,
    LogRecordsTextFilter,
)

FilterType = Union[
    LogRecordsBooleanFilter, LogRecordsDateFilter, LogRecordsIDFilter, LogRecordsNumberFilter, LogRecordsTextFilter
]



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.task_result_status import TaskResultStatus
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.validate_registered_scorer_result import ValidateRegisteredScorerResult


T = TypeVar("T", bound="RegisteredScorerTaskResultResponse")


@_attrs_define
class RegisteredScorerTaskResultResponse:
    """
    Attributes
    ----------
        id (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        status (TaskResultStatus):
        result (Union['ValidateRegisteredScorerResult', None, Unset, str]):
    """

    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    status: TaskResultStatus
    result: Union["ValidateRegisteredScorerResult", None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.validate_registered_scorer_result import ValidateRegisteredScorerResult

        id = self.id

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        status = self.status.value

        result: Union[None, Unset, dict[str, Any], str]
        if isinstance(self.result, Unset):
            result = UNSET
        elif isinstance(self.result, ValidateRegisteredScorerResult):
            result = self.result.to_dict()
        else:
            result = self.result

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"id": id, "created_at": created_at, "updated_at": updated_at, "status": status})
        if result is not UNSET:
            field_dict["result"] = result

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.validate_registered_scorer_result import ValidateRegisteredScorerResult

        d = dict(src_dict)
        id = d.pop("id")

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        status = TaskResultStatus(d.pop("status"))

        def _parse_result(data: object) -> Union["ValidateRegisteredScorerResult", None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return ValidateRegisteredScorerResult.from_dict(data)

            except:  # noqa: E722
                pass
            return cast(Union["ValidateRegisteredScorerResult", None, Unset, str], data)

        result = _parse_result(d.pop("result", UNSET))

        registered_scorer_task_result_response = cls(
            id=id, created_at=created_at, updated_at=updated_at, status=status, result=result
        )

        registered_scorer_task_result_response.additional_properties = d
        return registered_scorer_task_result_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.feedback_type import FeedbackType

T = TypeVar("T", bound="FeedbackRatingInfo")


@_attrs_define
class FeedbackRatingInfo:
    """
    Attributes
    ----------
        feedback_type (FeedbackType):
        value (Union[bool, int, list[str], str]):
        explanation (Union[None, str]):
    """

    feedback_type: FeedbackType
    value: Union[bool, int, list[str], str]
    explanation: Union[None, str]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        feedback_type = self.feedback_type.value

        value: Union[bool, int, list[str], str]
        value = self.value if isinstance(self.value, list) else self.value

        explanation: Union[None, str]
        explanation = self.explanation

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"feedback_type": feedback_type, "value": value, "explanation": explanation})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        feedback_type = FeedbackType(d.pop("feedback_type"))

        def _parse_value(data: object) -> Union[bool, int, list[str], str]:
            try:
                if not isinstance(data, list):
                    raise TypeError()
                return cast(list[str], data)

            except:  # noqa: E722
                pass
            return cast(Union[bool, int, list[str], str], data)

        value = _parse_value(d.pop("value"))

        def _parse_explanation(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        explanation = _parse_explanation(d.pop("explanation"))

        feedback_rating_info = cls(feedback_type=feedback_type, value=value, explanation=explanation)

        feedback_rating_info.additional_properties = d
        return feedback_rating_info

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import IntEnum


class TaskType(IntEnum):
    VALUE_0 = 0
    VALUE_1 = 1
    VALUE_2 = 2
    VALUE_3 = 3
    VALUE_4 = 4
    VALUE_5 = 5
    VALUE_6 = 6
    VALUE_7 = 7
    VALUE_8 = 8
    VALUE_9 = 9
    VALUE_10 = 10
    VALUE_11 = 11
    VALUE_12 = 12
    VALUE_13 = 13
    VALUE_14 = 14
    VALUE_15 = 15
    VALUE_16 = 16
    VALUE_17 = 17
    VALUE_18 = 18

    def __str__(self) -> str:
        return str(self.value)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.vertex_ai_integration_extra_type_0 import VertexAIIntegrationExtraType0


T = TypeVar("T", bound="VertexAIIntegration")


@_attrs_define
class VertexAIIntegration:
    """
    Attributes
    ----------
        id (Union[None, Unset, str]):
        name (Union[Literal['vertex_ai'], Unset]):  Default: 'vertex_ai'.
        extra (Union['VertexAIIntegrationExtraType0', None, Unset]):
    """

    id: Union[None, Unset, str] = UNSET
    name: Union[Literal["vertex_ai"], Unset] = "vertex_ai"
    extra: Union["VertexAIIntegrationExtraType0", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.vertex_ai_integration_extra_type_0 import VertexAIIntegrationExtraType0

        id: Union[None, Unset, str]
        id = UNSET if isinstance(self.id, Unset) else self.id

        name = self.name

        extra: Union[None, Unset, dict[str, Any]]
        if isinstance(self.extra, Unset):
            extra = UNSET
        elif isinstance(self.extra, VertexAIIntegrationExtraType0):
            extra = self.extra.to_dict()
        else:
            extra = self.extra

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if id is not UNSET:
            field_dict["id"] = id
        if name is not UNSET:
            field_dict["name"] = name
        if extra is not UNSET:
            field_dict["extra"] = extra

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.vertex_ai_integration_extra_type_0 import VertexAIIntegrationExtraType0

        d = dict(src_dict)

        def _parse_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        id = _parse_id(d.pop("id", UNSET))

        name = cast(Union[Literal["vertex_ai"], Unset], d.pop("name", UNSET))
        if name != "vertex_ai" and not isinstance(name, Unset):
            raise ValueError(f"name must match const 'vertex_ai', got '{name}'")

        def _parse_extra(data: object) -> Union["VertexAIIntegrationExtraType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return VertexAIIntegrationExtraType0.from_dict(data)

            except:  # noqa: E722
                pass
            return cast(Union["VertexAIIntegrationExtraType0", None, Unset], data)

        extra = _parse_extra(d.pop("extra", UNSET))

        vertex_ai_integration = cls(id=id, name=name, extra=extra)

        vertex_ai_integration.additional_properties = d
        return vertex_ai_integration

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""
Galileo wrapper for OpenAI that automatically logs prompts and responses.

This module provides a drop-in replacement for the OpenAI library that automatically
logs all prompts, responses, and related metadata to Galileo. It works by intercepting
calls to the OpenAI API and logging them using the Galileo logging system.

Note that the original OpenAI package is still required as a project dependency to use this wrapper.

Examples
--------
```python
# Import the wrapped OpenAI client instead of the original
from galileo.openai import openai

# Use it exactly as you would use the regular OpenAI client
response = openai.chat.completions.create(
    model="gpt-4o",
    messages=[
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Tell me about the solar system."}
    ]
)

# All prompts and responses are automatically logged to Galileo
print(response.choices[0].message.content)

# You can also use it with the galileo_context for more control
from galileo import galileo_context

with galileo_context(project="my-project", log_stream="my-log-stream"):
    response = openai.chat.completions.create(
        model="gpt-4o",
        messages=[
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "Tell me about the solar system."}
        ]
    )
```
"""

import logging
from typing import Any, Callable, Optional

import httpx
from wrapt import wrap_function_wrapper  # type: ignore[import-untyped]

from galileo.decorator import galileo_context
from galileo.logger import GalileoLogger
from galileo.openai.extractors import (
    OpenAiArgsExtractor,
    convert_to_galileo_message,
    extract_data_from_default_response,
    extract_input_data_from_kwargs,
    has_pending_function_calls,
    is_openai_v1,
    is_streaming_response,
    process_function_call_outputs,
    process_output_items,
)
from galileo.openai.models import OpenAiModuleDefinition
from galileo.openai.response_generator import ResponseGeneratorSync
from galileo.utils import _get_timestamp
from galileo.utils.serialization import serialize_to_str

try:
    import openai
    from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, NoneType, OpenAI  # noqa: F401
except ImportError:
    AsyncAzureOpenAI = None  # type: ignore[assignment]
    AsyncOpenAI = None  # type: ignore[assignment]
    AzureOpenAI = None  # type: ignore[assignment]
    OpenAI = None  # type: ignore[assignment]
    raise ModuleNotFoundError("Please install OpenAI to use this feature: 'pip install openai'")


_logger = logging.getLogger(__name__)


OPENAI_CLIENT_METHODS = [
    OpenAiModuleDefinition(
        module="openai.resources.chat.completions", object="Completions", method="create", type="chat", sync=True
    ),
    OpenAiModuleDefinition(
        module="openai.resources.responses", object="Responses", method="create", type="response", sync=True
    ),
    # Eventually add more OpenAI client library methods here
]


def _galileo_wrapper(func: Callable) -> Callable:
    def _with_galileo(open_ai_definitions: OpenAiModuleDefinition, initialize: Callable) -> Callable:
        def wrapper(wrapped: Callable, instance: Any, args: dict, kwargs: dict) -> Any:
            return func(open_ai_definitions, initialize, wrapped, args, kwargs)

        return wrapper

    return _with_galileo


@_galileo_wrapper
def _wrap(
    open_ai_resource: OpenAiModuleDefinition, initialize: Callable, wrapped: Callable, args: dict, kwargs: dict
) -> Any:
    start_time = _get_timestamp()
    arg_extractor = OpenAiArgsExtractor(*args, **kwargs)

    input_data = extract_input_data_from_kwargs(open_ai_resource, start_time, arg_extractor.get_galileo_args())

    galileo_logger: GalileoLogger = initialize()

    should_complete_trace = False
    if galileo_logger.current_parent():
        pass
    else:
        # If we don't have an active trace, start a new trace
        # We will conclude it at the end
        # convert to list of galileo messages since we can't send list of messages to span and want consistency
        if isinstance(input_data.input, list):
            trace_input_messages = [convert_to_galileo_message(msg) for msg in input_data.input]
        else:
            trace_input_messages = [convert_to_galileo_message(input_data.input)]

        # Serialize with "messages" wrapper for UI compatibility
        trace_input = {"messages": [msg.model_dump(exclude_none=True) for msg in trace_input_messages]}
        galileo_logger.start_trace(input=serialize_to_str(trace_input), name=input_data.name)
        should_complete_trace = True

    try:
        openai_response = None
        exc_info = None
        status_code = httpx.codes.OK
        try:
            openai_response = wrapped(**arg_extractor.get_openai_args())
        except openai.APIStatusError as exc:
            status_code = exc.status_code
            exc_info = exc

        if is_streaming_response(openai_response):
            # extract data from streaming response
            return ResponseGeneratorSync(
                resource=open_ai_resource,
                response=openai_response,
                input_data=input_data,
                logger=galileo_logger,
                should_complete_trace=should_complete_trace,
                status_code=status_code,
            )
        model, completion, usage = extract_data_from_default_response(
            open_ai_resource, (openai_response.__dict__ if openai_response and is_openai_v1() else openai_response)
        )

        if usage is None:
            usage = {}

        end_time = _get_timestamp()

        duration_ns = round((end_time - start_time).total_seconds() * 1e9)

        # convert to list of galileo messages since we can't send a regular list to span input
        if isinstance(input_data.input, list):
            span_input = [convert_to_galileo_message(msg) for msg in input_data.input]
        else:
            span_input = [convert_to_galileo_message(input_data.input)]

        # Process Responses API output items sequentially if present
        final_conversation_context = span_input.copy()
        output_items: list = []
        if open_ai_resource.type == "response" and openai_response:
            # First, process any function_call_output items in the input to create tool spans
            # This represents tool executions that happened before this API call
            if isinstance(input_data.input, list):
                process_function_call_outputs(input_data.input, galileo_logger)

            # Get output_items safely for Responses API
            # First try direct attribute access (works for Pydantic models)
            output_attr = getattr(openai_response, "output", None)
            if output_attr is not None:
                output_items = output_attr
            elif is_openai_v1() or hasattr(openai_response, "model_dump"):
                # Use model_dump() for Pydantic models
                response_mapping = openai_response.model_dump()
                output_items = response_mapping.get("output", [])
            else:
                # Fall back to __dict__ for dict-like responses
                output_items = openai_response.__dict__.get("output", [])

            # Process all output items sequentially and get the final context
            final_conversation_context = process_output_items(
                output_items,
                galileo_logger,
                model,
                span_input,
                input_data.model_parameters,
                status_code=status_code,
                tools=input_data.tools,
                usage=usage,
            )
        else:
            # For non-Responses API (chat or completion), create the main span as before
            span_output = convert_to_galileo_message(completion, "assistant")

            # Add a span to the current trace or span (if this is a nested trace)
            span = galileo_logger.add_llm_span(
                input=span_input,
                output=span_output,
                tools=input_data.tools,
                name=input_data.name,
                model=model,
                temperature=input_data.temperature,
                duration_ns=duration_ns,
                num_input_tokens=usage.get("input_tokens", 0),
                num_output_tokens=usage.get("output_tokens", 0),
                total_tokens=usage.get("total_tokens", 0),
                metadata={str(k): str(v) for k, v in input_data.model_parameters.items()},
                # openai client library doesn't return http_status code, so we only can hardcode it here
                # because we if we parsed and extracted data from response it means we get it and it's 200OK
                status_code=status_code,
            )
            span.metrics.num_reasoning_tokens = usage.get("reasoning_tokens", 0) if usage else 0
            span.metrics.num_cached_input_tokens = usage.get("cached_tokens", 0) if usage else 0

        # Conclude the trace if this is the top-level call
        # For Responses API: don't conclude if there are pending function calls (model waiting for tool results)
        has_pending_calls = (
            open_ai_resource.type == "response" and output_items and has_pending_function_calls(output_items)
        )

        if should_complete_trace and not has_pending_calls:
            if open_ai_resource.type == "response":
                # For Responses API, use the final conversation context from processing
                full_conversation = final_conversation_context
            else:
                # For other APIs, add the final span output
                full_conversation = []
                if isinstance(input_data.input, list):
                    full_conversation.extend([convert_to_galileo_message(msg) for msg in input_data.input])
                else:
                    full_conversation.append(convert_to_galileo_message(input_data.input))
                full_conversation.append(span_output)

            # Serialize with "messages" wrapper for UI compatibility
            trace_output = {"messages": [msg.model_dump(exclude_none=True) for msg in full_conversation]}
            galileo_logger.conclude(
                output=serialize_to_str(trace_output), duration_ns=duration_ns, status_code=status_code
            )

        # we want to re-raise exception after we process openai_response
        if exc_info:
            raise exc_info
        return openai_response
    except Exception as ex:
        _logger.error(f"Error while processing OpenAI request: {ex}")
        raise RuntimeError("Failed to process the OpenAI Request") from ex


class OpenAIGalileo:
    """
    This class is responsible for logging OpenAI API calls and logging them to Galileo.
    It wraps the OpenAI client methods to add logging functionality without changing
    the original API behavior.

    Attributes
    ----------
    _galileo_logger : Optional[GalileoLogger]
        The Galileo logger instance used for logging OpenAI API calls.
    """

    _galileo_logger: Optional[GalileoLogger] = None

    def initialize(self) -> Optional[GalileoLogger]:
        """
        Initialize a Galileo logger.

        Parameters
        ----------
        project : Optional[str]
            The project to log to. If None, uses the default project.
        log_stream : Optional[str]
            The log stream to log to. If None, uses the default log stream.

        Returns
        -------
        Optional[GalileoLogger]
            The initialized Galileo logger instance.
        """
        self._galileo_logger = galileo_context.get_logger_instance()

        return self._galileo_logger

    def register_tracing(self) -> None:
        """
        This method wraps the OpenAI client methods to intercept calls and log them to Galileo.
        It is called automatically when the module is imported.

        The wrapped methods include:
        - openai.resources.chat.completions.Completions.create

        Additional methods can be added to the OPENAI_CLIENT_METHODS list.
        """
        for resource in OPENAI_CLIENT_METHODS:
            wrap_function_wrapper(
                resource.module, f"{resource.object}.{resource.method}", (_wrap(resource, self.initialize))
            )


modifier = OpenAIGalileo()
modifier.register_tracing()



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.agentic_session_success_template_response_schema_type_0 import (
        AgenticSessionSuccessTemplateResponseSchemaType0,
    )
    from ..models.few_shot_example import FewShotExample


T = TypeVar("T", bound="AgenticSessionSuccessTemplate")


@_attrs_define
class AgenticSessionSuccessTemplate:
    r"""Template for the agentic session success metric,
    containing all the info necessary to send the agentic session success prompt.

    Attributes
    ----------
            metric_system_prompt (Union[Unset, str]):  Default: 'You will receive the complete chat history from a chatbot
                application between a user and an assistant.\n\nIn the chat history, the user will ask questions, which are
                answered with words, or make requests that require calling tools and resolving actions. Sometimes these are
                given as orders; treat them as if they were questions or requests. Each assistant turn may involve several steps
                that combine internal reflections, planning steps, selecting tools, and calling tools, and should always end
                with the assistant replying back to the user.\n\nYou will analyze the entire chat history and will respond back
                in the following JSON format:\n```json\n{\n    \\"all_user_asks\\": list[string],\n    \\"tasks\\":
                list[dict],\n    \\"ai_answered_all_asks\\": boolean,\n    \\"explanation\\": string\n}\n```\nwhere I will now
                explain how to populate each field.\n\n# Populating: all_user_asks\n\nPopulate `all_user_asks` with a list
                containing every user ask from the chat history. Review the chat history and generate a list with one entry for
                each user question, request, order, follow-up, clarification, etc. Ensure that every user ask is a separate
                item, even if this requires splitting the text mid-sentence. Each item should include enough context to be
                understandable on its own. It is acceptable to have shared context between items and to incorporate parts of
                sentences as needed.\n\n# Populating: Tasks\n\nThis is the most complex field to populate. You will write a JSON
                array where each element is called a task and follows the schema:\n\n```json\n{\n    \\"initial_user_ask\\":
                string,\n    \\"user_ask_refinements\\": list[string],\n    \\"final_user_ask\\": string,\n
                \\"direct_answer\\": string,\n    \\"indirect_answer\\": string,\n    \\"tools_input_output\\": list[string],\n
                \\"properties\\" : {\n        \\"coherent\\": boolean,\n        \\"factually_correct\\": boolean,\n
                \\"comprehensively_answers_final_user_ask\\": boolean,\n        \\"does_not_contradict_tools_output\\":
                boolean,\n        \\"tools_output_summary_is_accurate\\": boolean,\n    },\n    \\"boolean_properties\\":
                list[boolean],\n    \\"answer_satisfies_properties\\": boolean\n}\n```\n\nThe high-level goal is to list all
                tasks and their resolutions and to determine whether each task has been successfully accomplished.\n\n## Step 1:
                initial_user_ask, user_ask_refinements and final_user_ask\n\nFirst, identify the `initial_user_ask` that starts
                the task, as well as any `user_ask_refinements` related to the same task. To do this, first loop through the
                entries in `all_user_asks`. If an entry already appears in a previous task, ignore it; otherwise, consider it as
                the `initial_user_ask`. Next, examine the remaining entries in `all_user_asks` and fill `user_ask_refinements`
                with all those related to the `initial_user_ask`, meaning they either refine it or continue the same
                ask.\n\nFinally, create a coherent `final_user_ask` containing the most updated version of the ask by starting
                with the initial one and incorporating or replacing any parts with their refinements. This will be the ask that
                the assistant will attempt to answer.\n\n## Step 2: direct_answer and indirect_answer\n\nExtract every direct
                and indirect answer that responds to the `final_user_ask`.\n\nAn indirect answer is a part of the assistant\'s
                reponse that tries to respond to `final_user_ask` and satisfies any of the following:\n- it mentions limitations
                or the inability to complete the `final_user_ask`,\n- it references a failed attempt to complete the
                `final_user_ask`,\n- it suggests offering help with a different ask than the `final_user_ask`,\n- it requests
                further information or clarifications from the user.\nAdd any piece of the assistant\'s response looking like an
                indirect answer to `indirect_answer`.\n\nA direct answer is a part of an assistant\'s response that either:\n-
                directly responds to the `final_user_ask`,\n- confirms a successful resolution of the `final_user_ask`.\nIf
                there are multiple direct answers, simply concatenate them into a longer answer. If there are no direct answers
                satisfying the above conditions, leave the field `direct_answer` empty.\n\nNote that a piece of an answer cannot
                be both direct and indirect, you should pick the field in which to add it.\n\n## Step 3:
                tools_input_output\n\nIf `direct_answer` is empty, skip this step.\n\nExamine each assistant step and identify
                which tool or function output seemingly contributed to creating any part of the answer from `direct_answer`. If
                an assistant step immediately before or after the tool call mentions using or having used the tool for answering
                the `final_user_ask`, the tool call should be associated with this ask. Additionally, if any part of the answer
                closely aligns with the output of a tool, the tool call should also be associated with this ask.\n\nCreate a
                list containing the concatenated input and output of each tool used in formulating any part of the answer from
                `direct_answer`. The tool input is noted as an assistant step before calling the tool, and the tool output is
                recorded as a tool step.\n\n## Step 4: properties, boolean_properties and answer_satisfies_properties\n\nIf
                `direct_answer` is empty, set every boolean in `properties`, `boolean_properties` and
                `answer_satisfies_properties` to `false`.\n\nFor each part of the answer from `direct_answer`, evaluate the
                following properties one by one to determine which are satisfied and which are not:\n\n- **coherent**: The
                answer is coherent with itself and does not contain internal contradictions.\n- **factually_correct**: The parts
                of the answer that do not come from the output of a tool are factually correct.\n-
                **comprehensively_answers_final_user_ask**: The answer specifically responds to the `final_user_ask`, carefully
                addressing every aspect of the ask without deviation or omission, ensuring that no details or parts of the ask
                are left unanswered.\n- **does_not_contradict_tools_output**: No citation of a tool\'s output contradict any
                text from `tools_input_output`.\n- **tools_output_summary_is_accurate**: Every summary of a tool\'s output is
                accurate with the tool\'s output from `tools_input_output`. In particular it does not omit critical information
                relevant to the `final_user_ask` and does not contain made-up information.\n\nAfter assessing each of these
                properties, copy the resulting boolean values into the list `boolean_properties`.\n\nFinally, set
                `answer_satisfies_properties` to `false` if any entry in `boolean_properties` is set to `false`; otherwise, set
                `answer_satisfies_properties` to `true`.\n\n# Populating: ai_answered_all_asks\n\nRespond `true` if every task
                has `answer_satisfies_properties` set to `true`, otherwise respond `false`. If `all_user_asks` is empty, set
                `answer_satisfies_properties` to `true`.\n\n# Populating: explanation\n\nIf any user ask has
                `answer_satisfies_properties` set to `false`, explain why it didn\'t satisfy all the properties. Otherwise
                summarize in a few words each ask and the provided answer.\n\nIf `all_user_asks` is empty, mention that you did
                not find any user ask. If `direct_answer` is empty, mention that no resultion to the `final_user_ask` was
                provided.\n\nYou must respond with a valid JSON object; be sure to escape special characters.'.
            metric_description (Union[Unset, str]):  Default: 'I have a multi-turn chatbot application where the assistant
                is an agent that has access to tools. I want a metric that assesses whether the session should be considered
                successful, in the sense that the assistant fully answered or resolved all user queries and requests.'.
            value_field_name (Union[Unset, str]):  Default: 'ai_answered_all_asks'.
            explanation_field_name (Union[Unset, str]): Field name to look for in the chainpoll response, for the
                explanation. Default: 'explanation'.
            template (Union[Unset, str]):  Default: 'Here is a the chatbot history:\n```\n{query}\n```\nNow perform the
                evaluation on the chat history as described in the system prompt.'.
            metric_few_shot_examples (Union[Unset, list['FewShotExample']]):
            response_schema (Union['AgenticSessionSuccessTemplateResponseSchemaType0', None, Unset]): Response schema for
                the output
    """

    metric_system_prompt: Union[Unset, str] = (
        'You will receive the complete chat history from a chatbot application between a user and an assistant.\n\nIn the chat history, the user will ask questions, which are answered with words, or make requests that require calling tools and resolving actions. Sometimes these are given as orders; treat them as if they were questions or requests. Each assistant turn may involve several steps that combine internal reflections, planning steps, selecting tools, and calling tools, and should always end with the assistant replying back to the user.\n\nYou will analyze the entire chat history and will respond back in the following JSON format:\n```json\n{\n    \\"all_user_asks\\": list[string],\n    \\"tasks\\": list[dict],\n    \\"ai_answered_all_asks\\": boolean,\n    \\"explanation\\": string\n}\n```\nwhere I will now explain how to populate each field.\n\n# Populating: all_user_asks\n\nPopulate `all_user_asks` with a list containing every user ask from the chat history. Review the chat history and generate a list with one entry for each user question, request, order, follow-up, clarification, etc. Ensure that every user ask is a separate item, even if this requires splitting the text mid-sentence. Each item should include enough context to be understandable on its own. It is acceptable to have shared context between items and to incorporate parts of sentences as needed.\n\n# Populating: Tasks\n\nThis is the most complex field to populate. You will write a JSON array where each element is called a task and follows the schema:\n\n```json\n{\n    \\"initial_user_ask\\": string,\n    \\"user_ask_refinements\\": list[string],\n    \\"final_user_ask\\": string,\n    \\"direct_answer\\": string,\n    \\"indirect_answer\\": string,\n    \\"tools_input_output\\": list[string],\n    \\"properties\\" : {\n        \\"coherent\\": boolean,\n        \\"factually_correct\\": boolean,\n        \\"comprehensively_answers_final_user_ask\\": boolean,\n        \\"does_not_contradict_tools_output\\": boolean,\n        \\"tools_output_summary_is_accurate\\": boolean,\n    },\n    \\"boolean_properties\\": list[boolean],\n    \\"answer_satisfies_properties\\": boolean\n}\n```\n\nThe high-level goal is to list all tasks and their resolutions and to determine whether each task has been successfully accomplished.\n\n## Step 1: initial_user_ask, user_ask_refinements and final_user_ask\n\nFirst, identify the `initial_user_ask` that starts the task, as well as any `user_ask_refinements` related to the same task. To do this, first loop through the entries in `all_user_asks`. If an entry already appears in a previous task, ignore it; otherwise, consider it as the `initial_user_ask`. Next, examine the remaining entries in `all_user_asks` and fill `user_ask_refinements` with all those related to the `initial_user_ask`, meaning they either refine it or continue the same ask.\n\nFinally, create a coherent `final_user_ask` containing the most updated version of the ask by starting with the initial one and incorporating or replacing any parts with their refinements. This will be the ask that the assistant will attempt to answer.\n\n## Step 2: direct_answer and indirect_answer\n\nExtract every direct and indirect answer that responds to the `final_user_ask`.\n\nAn indirect answer is a part of the assistant\'s reponse that tries to respond to `final_user_ask` and satisfies any of the following:\n- it mentions limitations or the inability to complete the `final_user_ask`,\n- it references a failed attempt to complete the `final_user_ask`,\n- it suggests offering help with a different ask than the `final_user_ask`,\n- it requests further information or clarifications from the user.\nAdd any piece of the assistant\'s response looking like an indirect answer to `indirect_answer`.\n\nA direct answer is a part of an assistant\'s response that either:\n- directly responds to the `final_user_ask`,\n- confirms a successful resolution of the `final_user_ask`.\nIf there are multiple direct answers, simply concatenate them into a longer answer. If there are no direct answers satisfying the above conditions, leave the field `direct_answer` empty.\n\nNote that a piece of an answer cannot be both direct and indirect, you should pick the field in which to add it.\n\n## Step 3: tools_input_output\n\nIf `direct_answer` is empty, skip this step.\n\nExamine each assistant step and identify which tool or function output seemingly contributed to creating any part of the answer from `direct_answer`. If an assistant step immediately before or after the tool call mentions using or having used the tool for answering the `final_user_ask`, the tool call should be associated with this ask. Additionally, if any part of the answer closely aligns with the output of a tool, the tool call should also be associated with this ask.\n\nCreate a list containing the concatenated input and output of each tool used in formulating any part of the answer from `direct_answer`. The tool input is noted as an assistant step before calling the tool, and the tool output is recorded as a tool step.\n\n## Step 4: properties, boolean_properties and answer_satisfies_properties\n\nIf `direct_answer` is empty, set every boolean in `properties`, `boolean_properties` and `answer_satisfies_properties` to `false`.\n\nFor each part of the answer from `direct_answer`, evaluate the following properties one by one to determine which are satisfied and which are not:\n\n- **coherent**: The answer is coherent with itself and does not contain internal contradictions.\n- **factually_correct**: The parts of the answer that do not come from the output of a tool are factually correct.\n- **comprehensively_answers_final_user_ask**: The answer specifically responds to the `final_user_ask`, carefully addressing every aspect of the ask without deviation or omission, ensuring that no details or parts of the ask are left unanswered.\n- **does_not_contradict_tools_output**: No citation of a tool\'s output contradict any text from `tools_input_output`.\n- **tools_output_summary_is_accurate**: Every summary of a tool\'s output is accurate with the tool\'s output from `tools_input_output`. In particular it does not omit critical information relevant to the `final_user_ask` and does not contain made-up information.\n\nAfter assessing each of these properties, copy the resulting boolean values into the list `boolean_properties`.\n\nFinally, set `answer_satisfies_properties` to `false` if any entry in `boolean_properties` is set to `false`; otherwise, set `answer_satisfies_properties` to `true`.\n\n# Populating: ai_answered_all_asks\n\nRespond `true` if every task has `answer_satisfies_properties` set to `true`, otherwise respond `false`. If `all_user_asks` is empty, set `answer_satisfies_properties` to `true`.\n\n# Populating: explanation\n\nIf any user ask has `answer_satisfies_properties` set to `false`, explain why it didn\'t satisfy all the properties. Otherwise summarize in a few words each ask and the provided answer.\n\nIf `all_user_asks` is empty, mention that you did not find any user ask. If `direct_answer` is empty, mention that no resultion to the `final_user_ask` was provided.\n\nYou must respond with a valid JSON object; be sure to escape special characters.'
    )
    metric_description: Union[Unset, str] = (
        "I have a multi-turn chatbot application where the assistant is an agent that has access to tools. I want a metric that assesses whether the session should be considered successful, in the sense that the assistant fully answered or resolved all user queries and requests."
    )
    value_field_name: Union[Unset, str] = "ai_answered_all_asks"
    explanation_field_name: Union[Unset, str] = "explanation"
    template: Union[Unset, str] = (
        "Here is a the chatbot history:\n```\n{query}\n```\nNow perform the evaluation on the chat history as described in the system prompt."
    )
    metric_few_shot_examples: Union[Unset, list["FewShotExample"]] = UNSET
    response_schema: Union["AgenticSessionSuccessTemplateResponseSchemaType0", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.agentic_session_success_template_response_schema_type_0 import (
            AgenticSessionSuccessTemplateResponseSchemaType0,
        )

        metric_system_prompt = self.metric_system_prompt

        metric_description = self.metric_description

        value_field_name = self.value_field_name

        explanation_field_name = self.explanation_field_name

        template = self.template

        metric_few_shot_examples: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.metric_few_shot_examples, Unset):
            metric_few_shot_examples = []
            for metric_few_shot_examples_item_data in self.metric_few_shot_examples:
                metric_few_shot_examples_item = metric_few_shot_examples_item_data.to_dict()
                metric_few_shot_examples.append(metric_few_shot_examples_item)

        response_schema: Union[None, Unset, dict[str, Any]]
        if isinstance(self.response_schema, Unset):
            response_schema = UNSET
        elif isinstance(self.response_schema, AgenticSessionSuccessTemplateResponseSchemaType0):
            response_schema = self.response_schema.to_dict()
        else:
            response_schema = self.response_schema

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if metric_system_prompt is not UNSET:
            field_dict["metric_system_prompt"] = metric_system_prompt
        if metric_description is not UNSET:
            field_dict["metric_description"] = metric_description
        if value_field_name is not UNSET:
            field_dict["value_field_name"] = value_field_name
        if explanation_field_name is not UNSET:
            field_dict["explanation_field_name"] = explanation_field_name
        if template is not UNSET:
            field_dict["template"] = template
        if metric_few_shot_examples is not UNSET:
            field_dict["metric_few_shot_examples"] = metric_few_shot_examples
        if response_schema is not UNSET:
            field_dict["response_schema"] = response_schema

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.agentic_session_success_template_response_schema_type_0 import (
            AgenticSessionSuccessTemplateResponseSchemaType0,
        )
        from ..models.few_shot_example import FewShotExample

        d = dict(src_dict)
        metric_system_prompt = d.pop("metric_system_prompt", UNSET)

        metric_description = d.pop("metric_description", UNSET)

        value_field_name = d.pop("value_field_name", UNSET)

        explanation_field_name = d.pop("explanation_field_name", UNSET)

        template = d.pop("template", UNSET)

        metric_few_shot_examples = []
        _metric_few_shot_examples = d.pop("metric_few_shot_examples", UNSET)
        for metric_few_shot_examples_item_data in _metric_few_shot_examples or []:
            metric_few_shot_examples_item = FewShotExample.from_dict(metric_few_shot_examples_item_data)

            metric_few_shot_examples.append(metric_few_shot_examples_item)

        def _parse_response_schema(
            data: object,
        ) -> Union["AgenticSessionSuccessTemplateResponseSchemaType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return AgenticSessionSuccessTemplateResponseSchemaType0.from_dict(data)

            except:  # noqa: E722
                pass
            return cast(Union["AgenticSessionSuccessTemplateResponseSchemaType0", None, Unset], data)

        response_schema = _parse_response_schema(d.pop("response_schema", UNSET))

        agentic_session_success_template = cls(
            metric_system_prompt=metric_system_prompt,
            metric_description=metric_description,
            value_field_name=value_field_name,
            explanation_field_name=explanation_field_name,
            template=template,
            metric_few_shot_examples=metric_few_shot_examples,
            response_schema=response_schema,
        )

        agentic_session_success_template.additional_properties = d
        return agentic_session_success_template

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.and_node_log_records_filter import AndNodeLogRecordsFilter
    from ..models.filter_leaf_log_records_filter import FilterLeafLogRecordsFilter
    from ..models.or_node_log_records_filter import OrNodeLogRecordsFilter


T = TypeVar("T", bound="NotNodeLogRecordsFilter")


@_attrs_define
class NotNodeLogRecordsFilter:
    """
    Attributes
    ----------
        not_ (Union['AndNodeLogRecordsFilter', 'FilterLeafLogRecordsFilter', 'NotNodeLogRecordsFilter',
            'OrNodeLogRecordsFilter']):
    """

    not_: Union[
        "AndNodeLogRecordsFilter", "FilterLeafLogRecordsFilter", "NotNodeLogRecordsFilter", "OrNodeLogRecordsFilter"
    ]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.and_node_log_records_filter import AndNodeLogRecordsFilter
        from ..models.filter_leaf_log_records_filter import FilterLeafLogRecordsFilter
        from ..models.or_node_log_records_filter import OrNodeLogRecordsFilter

        not_: dict[str, Any]
        if isinstance(self.not_, (FilterLeafLogRecordsFilter, AndNodeLogRecordsFilter, OrNodeLogRecordsFilter)):
            not_ = self.not_.to_dict()
        else:
            not_ = self.not_.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"not": not_})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.and_node_log_records_filter import AndNodeLogRecordsFilter
        from ..models.filter_leaf_log_records_filter import FilterLeafLogRecordsFilter
        from ..models.or_node_log_records_filter import OrNodeLogRecordsFilter

        d = dict(src_dict)

        def _parse_not_(
            data: object,
        ) -> Union[
            "AndNodeLogRecordsFilter", "FilterLeafLogRecordsFilter", "NotNodeLogRecordsFilter", "OrNodeLogRecordsFilter"
        ]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return FilterLeafLogRecordsFilter.from_dict(data)

            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return AndNodeLogRecordsFilter.from_dict(data)

            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return OrNodeLogRecordsFilter.from_dict(data)

            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            return NotNodeLogRecordsFilter.from_dict(data)

        not_ = _parse_not_(d.pop("not"))

        not_node_log_records_filter = cls(not_=not_)

        not_node_log_records_filter.additional_properties = d
        return not_node_log_records_filter

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# mypy: disable-error-code=syntax
# We need to ignore syntax errors until https://github.com/python/mypy/issues/17535 is resolved.
from typing import Any, ClassVar, Optional

from pydantic_core import Url

from galileo_core.schemas.base_config import GalileoConfig


class GalileoPythonConfig(GalileoConfig):
    # Config file for this project.
    config_filename: str = "galileo-python-config.json"
    console_url: Url = "https://app.galileo.ai"

    _instance: ClassVar[Optional["GalileoPythonConfig"]] = None

    def reset(self) -> None:
        GalileoPythonConfig._instance = None
        super().reset()

    @classmethod
    def get(cls, **kwargs: Any) -> "GalileoPythonConfig":
        cls._instance = cls._get(cls._instance, **kwargs)
        assert cls._instance is not None, "Failed to initialize GalileoPythonConfig"
        return cls._instance



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.document import Document
    from ..models.extended_agent_span_record_with_children import ExtendedAgentSpanRecordWithChildren
    from ..models.extended_llm_span_record import ExtendedLlmSpanRecord
    from ..models.extended_retriever_span_record_with_children_dataset_metadata import (
        ExtendedRetrieverSpanRecordWithChildrenDatasetMetadata,
    )
    from ..models.extended_retriever_span_record_with_children_feedback_rating_info import (
        ExtendedRetrieverSpanRecordWithChildrenFeedbackRatingInfo,
    )
    from ..models.extended_retriever_span_record_with_children_metric_info_type_0 import (
        ExtendedRetrieverSpanRecordWithChildrenMetricInfoType0,
    )
    from ..models.extended_retriever_span_record_with_children_user_metadata import (
        ExtendedRetrieverSpanRecordWithChildrenUserMetadata,
    )
    from ..models.extended_tool_span_record_with_children import ExtendedToolSpanRecordWithChildren
    from ..models.extended_workflow_span_record_with_children import ExtendedWorkflowSpanRecordWithChildren
    from ..models.metrics import Metrics


T = TypeVar("T", bound="ExtendedRetrieverSpanRecordWithChildren")


@_attrs_define
class ExtendedRetrieverSpanRecordWithChildren:
    """
    Attributes
    ----------
        id (str): Galileo ID of the session, trace or span
        session_id (str): Galileo ID of the session containing the trace (or the same value as id for a trace)
        project_id (str): Galileo ID of the project associated with this trace or span
        run_id (str): Galileo ID of the run (log stream or experiment) associated with this trace or span
        parent_id (str): Galileo ID of the parent of this span
        spans (Union[Unset, list[Union['ExtendedAgentSpanRecordWithChildren', 'ExtendedLlmSpanRecord',
            'ExtendedRetrieverSpanRecordWithChildren', 'ExtendedToolSpanRecordWithChildren',
            'ExtendedWorkflowSpanRecordWithChildren']]]):
        type_ (Union[Literal['retriever'], Unset]): Type of the trace, span or session. Default: 'retriever'.
        input_ (Union[Unset, str]): Input to the trace or span. Default: ''.
        redacted_input (Union[None, Unset, str]): Redacted input of the trace or span.
        output (Union[Unset, list['Document']]): Output of the trace or span.
        redacted_output (Union[None, Unset, list['Document']]): Redacted output of the trace or span.
        name (Union[Unset, str]): Name of the trace, span or session. Default: ''.
        created_at (Union[Unset, datetime.datetime]): Timestamp of the trace or span's creation.
        user_metadata (Union[Unset, ExtendedRetrieverSpanRecordWithChildrenUserMetadata]): Metadata associated with this
            trace or span.
        tags (Union[Unset, list[str]]): Tags associated with this trace or span.
        status_code (Union[None, Unset, int]): Status code of the trace or span. Used for logging failure or error
            states.
        metrics (Union[Unset, Metrics]):
        external_id (Union[None, Unset, str]): A user-provided session, trace or span ID.
        dataset_input (Union[None, Unset, str]): Input to the dataset associated with this trace
        dataset_output (Union[None, Unset, str]): Output from the dataset associated with this trace
        dataset_metadata (Union[Unset, ExtendedRetrieverSpanRecordWithChildrenDatasetMetadata]): Metadata from the
            dataset associated with this trace
        trace_id (Union[None, Unset, str]): Galileo ID of the trace containing the span (or the same value as id for a
            trace)
        updated_at (Union[None, Unset, datetime.datetime]): Timestamp of the session or trace or span's last update
        has_children (Union[None, Unset, bool]): Whether or not this trace or span has child spans
        metrics_batch_id (Union[None, Unset, str]): Galileo ID of the metrics batch associated with this trace or span
        session_batch_id (Union[None, Unset, str]): Galileo ID of the metrics batch associated with this trace or span
        feedback_rating_info (Union[Unset, ExtendedRetrieverSpanRecordWithChildrenFeedbackRatingInfo]): Feedback
            information related to the record
        metric_info (Union['ExtendedRetrieverSpanRecordWithChildrenMetricInfoType0', None, Unset]): Detailed information
            about the metrics associated with this trace or span
        is_complete (Union[Unset, bool]): Whether the parent trace is complete or not Default: True.
        step_number (Union[None, Unset, int]): Topological step number of the span.
    """

    id: str
    session_id: str
    project_id: str
    run_id: str
    parent_id: str
    spans: Union[
        Unset,
        list[
            Union[
                "ExtendedAgentSpanRecordWithChildren",
                "ExtendedLlmSpanRecord",
                "ExtendedRetrieverSpanRecordWithChildren",
                "ExtendedToolSpanRecordWithChildren",
                "ExtendedWorkflowSpanRecordWithChildren",
            ]
        ],
    ] = UNSET
    type_: Union[Literal["retriever"], Unset] = "retriever"
    input_: Union[Unset, str] = ""
    redacted_input: Union[None, Unset, str] = UNSET
    output: Union[Unset, list["Document"]] = UNSET
    redacted_output: Union[None, Unset, list["Document"]] = UNSET
    name: Union[Unset, str] = ""
    created_at: Union[Unset, datetime.datetime] = UNSET
    user_metadata: Union[Unset, "ExtendedRetrieverSpanRecordWithChildrenUserMetadata"] = UNSET
    tags: Union[Unset, list[str]] = UNSET
    status_code: Union[None, Unset, int] = UNSET
    metrics: Union[Unset, "Metrics"] = UNSET
    external_id: Union[None, Unset, str] = UNSET
    dataset_input: Union[None, Unset, str] = UNSET
    dataset_output: Union[None, Unset, str] = UNSET
    dataset_metadata: Union[Unset, "ExtendedRetrieverSpanRecordWithChildrenDatasetMetadata"] = UNSET
    trace_id: Union[None, Unset, str] = UNSET
    updated_at: Union[None, Unset, datetime.datetime] = UNSET
    has_children: Union[None, Unset, bool] = UNSET
    metrics_batch_id: Union[None, Unset, str] = UNSET
    session_batch_id: Union[None, Unset, str] = UNSET
    feedback_rating_info: Union[Unset, "ExtendedRetrieverSpanRecordWithChildrenFeedbackRatingInfo"] = UNSET
    metric_info: Union["ExtendedRetrieverSpanRecordWithChildrenMetricInfoType0", None, Unset] = UNSET
    is_complete: Union[Unset, bool] = True
    step_number: Union[None, Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.extended_agent_span_record_with_children import ExtendedAgentSpanRecordWithChildren
        from ..models.extended_llm_span_record import ExtendedLlmSpanRecord
        from ..models.extended_retriever_span_record_with_children_metric_info_type_0 import (
            ExtendedRetrieverSpanRecordWithChildrenMetricInfoType0,
        )
        from ..models.extended_tool_span_record_with_children import ExtendedToolSpanRecordWithChildren
        from ..models.extended_workflow_span_record_with_children import ExtendedWorkflowSpanRecordWithChildren

        id = self.id

        session_id = self.session_id

        project_id = self.project_id

        run_id = self.run_id

        parent_id = self.parent_id

        spans: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.spans, Unset):
            spans = []
            for spans_item_data in self.spans:
                spans_item: dict[str, Any]
                if isinstance(
                    spans_item_data,
                    (
                        ExtendedAgentSpanRecordWithChildren,
                        ExtendedWorkflowSpanRecordWithChildren,
                        ExtendedLlmSpanRecord,
                        ExtendedToolSpanRecordWithChildren,
                    ),
                ):
                    spans_item = spans_item_data.to_dict()
                else:
                    spans_item = spans_item_data.to_dict()

                spans.append(spans_item)

        type_ = self.type_

        input_ = self.input_

        redacted_input: Union[None, Unset, str]
        redacted_input = UNSET if isinstance(self.redacted_input, Unset) else self.redacted_input

        output: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.output, Unset):
            output = []
            for output_item_data in self.output:
                output_item = output_item_data.to_dict()
                output.append(output_item)

        redacted_output: Union[None, Unset, list[dict[str, Any]]]
        if isinstance(self.redacted_output, Unset):
            redacted_output = UNSET
        elif isinstance(self.redacted_output, list):
            redacted_output = []
            for redacted_output_type_0_item_data in self.redacted_output:
                redacted_output_type_0_item = redacted_output_type_0_item_data.to_dict()
                redacted_output.append(redacted_output_type_0_item)

        else:
            redacted_output = self.redacted_output

        name = self.name

        created_at: Union[Unset, str] = UNSET
        if not isinstance(self.created_at, Unset):
            created_at = self.created_at.isoformat()

        user_metadata: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.user_metadata, Unset):
            user_metadata = self.user_metadata.to_dict()

        tags: Union[Unset, list[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        status_code: Union[None, Unset, int]
        status_code = UNSET if isinstance(self.status_code, Unset) else self.status_code

        metrics: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.metrics, Unset):
            metrics = self.metrics.to_dict()

        external_id: Union[None, Unset, str]
        external_id = UNSET if isinstance(self.external_id, Unset) else self.external_id

        dataset_input: Union[None, Unset, str]
        dataset_input = UNSET if isinstance(self.dataset_input, Unset) else self.dataset_input

        dataset_output: Union[None, Unset, str]
        dataset_output = UNSET if isinstance(self.dataset_output, Unset) else self.dataset_output

        dataset_metadata: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.dataset_metadata, Unset):
            dataset_metadata = self.dataset_metadata.to_dict()

        trace_id: Union[None, Unset, str]
        trace_id = UNSET if isinstance(self.trace_id, Unset) else self.trace_id

        updated_at: Union[None, Unset, str]
        if isinstance(self.updated_at, Unset):
            updated_at = UNSET
        elif isinstance(self.updated_at, datetime.datetime):
            updated_at = self.updated_at.isoformat()
        else:
            updated_at = self.updated_at

        has_children: Union[None, Unset, bool]
        has_children = UNSET if isinstance(self.has_children, Unset) else self.has_children

        metrics_batch_id: Union[None, Unset, str]
        metrics_batch_id = UNSET if isinstance(self.metrics_batch_id, Unset) else self.metrics_batch_id

        session_batch_id: Union[None, Unset, str]
        session_batch_id = UNSET if isinstance(self.session_batch_id, Unset) else self.session_batch_id

        feedback_rating_info: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.feedback_rating_info, Unset):
            feedback_rating_info = self.feedback_rating_info.to_dict()

        metric_info: Union[None, Unset, dict[str, Any]]
        if isinstance(self.metric_info, Unset):
            metric_info = UNSET
        elif isinstance(self.metric_info, ExtendedRetrieverSpanRecordWithChildrenMetricInfoType0):
            metric_info = self.metric_info.to_dict()
        else:
            metric_info = self.metric_info

        is_complete = self.is_complete

        step_number: Union[None, Unset, int]
        step_number = UNSET if isinstance(self.step_number, Unset) else self.step_number

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {"id": id, "session_id": session_id, "project_id": project_id, "run_id": run_id, "parent_id": parent_id}
        )
        if spans is not UNSET:
            field_dict["spans"] = spans
        if type_ is not UNSET:
            field_dict["type"] = type_
        if input_ is not UNSET:
            field_dict["input"] = input_
        if redacted_input is not UNSET:
            field_dict["redacted_input"] = redacted_input
        if output is not UNSET:
            field_dict["output"] = output
        if redacted_output is not UNSET:
            field_dict["redacted_output"] = redacted_output
        if name is not UNSET:
            field_dict["name"] = name
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if user_metadata is not UNSET:
            field_dict["user_metadata"] = user_metadata
        if tags is not UNSET:
            field_dict["tags"] = tags
        if status_code is not UNSET:
            field_dict["status_code"] = status_code
        if metrics is not UNSET:
            field_dict["metrics"] = metrics
        if external_id is not UNSET:
            field_dict["external_id"] = external_id
        if dataset_input is not UNSET:
            field_dict["dataset_input"] = dataset_input
        if dataset_output is not UNSET:
            field_dict["dataset_output"] = dataset_output
        if dataset_metadata is not UNSET:
            field_dict["dataset_metadata"] = dataset_metadata
        if trace_id is not UNSET:
            field_dict["trace_id"] = trace_id
        if updated_at is not UNSET:
            field_dict["updated_at"] = updated_at
        if has_children is not UNSET:
            field_dict["has_children"] = has_children
        if metrics_batch_id is not UNSET:
            field_dict["metrics_batch_id"] = metrics_batch_id
        if session_batch_id is not UNSET:
            field_dict["session_batch_id"] = session_batch_id
        if feedback_rating_info is not UNSET:
            field_dict["feedback_rating_info"] = feedback_rating_info
        if metric_info is not UNSET:
            field_dict["metric_info"] = metric_info
        if is_complete is not UNSET:
            field_dict["is_complete"] = is_complete
        if step_number is not UNSET:
            field_dict["step_number"] = step_number

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.document import Document
        from ..models.extended_agent_span_record_with_children import ExtendedAgentSpanRecordWithChildren
        from ..models.extended_retriever_span_record_with_children_dataset_metadata import (
            ExtendedRetrieverSpanRecordWithChildrenDatasetMetadata,
        )
        from ..models.extended_retriever_span_record_with_children_feedback_rating_info import (
            ExtendedRetrieverSpanRecordWithChildrenFeedbackRatingInfo,
        )
        from ..models.extended_retriever_span_record_with_children_metric_info_type_0 import (
            ExtendedRetrieverSpanRecordWithChildrenMetricInfoType0,
        )
        from ..models.extended_retriever_span_record_with_children_user_metadata import (
            ExtendedRetrieverSpanRecordWithChildrenUserMetadata,
        )
        from ..models.extended_tool_span_record_with_children import ExtendedToolSpanRecordWithChildren
        from ..models.extended_workflow_span_record_with_children import ExtendedWorkflowSpanRecordWithChildren
        from ..models.metrics import Metrics

        d = dict(src_dict)
        id = d.pop("id")

        session_id = d.pop("session_id")

        project_id = d.pop("project_id")

        run_id = d.pop("run_id")

        parent_id = d.pop("parent_id")

        spans = []
        _spans = d.pop("spans", UNSET)
        for spans_item_data in _spans or []:

            def _parse_spans_item(
                data: object,
            ) -> Union[
                "ExtendedAgentSpanRecordWithChildren",
                "ExtendedLlmSpanRecord",
                "ExtendedRetrieverSpanRecordWithChildren",
                "ExtendedToolSpanRecordWithChildren",
                "ExtendedWorkflowSpanRecordWithChildren",
            ]:
                # Discriminator-aware parsing for Extended*Record types
                if isinstance(data, dict) and "type" in data:
                    type_value = data.get("type")

                    # Hardcoded discriminator mapping for Extended*Record types
                    if type_value == "trace":
                        try:
                            from ..models.extended_trace_record import ExtendedTraceRecord

                            return ExtendedTraceRecord.from_dict(data)
                        except:  # noqa: E722
                            pass
                    elif type_value == "agent":
                        try:
                            from ..models.extended_agent_span_record import ExtendedAgentSpanRecord

                            return ExtendedAgentSpanRecord.from_dict(data)
                        except:  # noqa: E722
                            pass
                    elif type_value == "workflow":
                        try:
                            from ..models.extended_workflow_span_record import ExtendedWorkflowSpanRecord

                            return ExtendedWorkflowSpanRecord.from_dict(data)
                        except:  # noqa: E722
                            pass
                    elif type_value == "llm":
                        try:
                            from ..models.extended_llm_span_record import ExtendedLlmSpanRecord

                            return ExtendedLlmSpanRecord.from_dict(data)
                        except:  # noqa: E722
                            pass
                    elif type_value == "tool":
                        try:
                            from ..models.extended_tool_span_record import ExtendedToolSpanRecord

                            return ExtendedToolSpanRecord.from_dict(data)
                        except:  # noqa: E722
                            pass
                    elif type_value == "retriever":
                        try:
                            from ..models.extended_retriever_span_record import ExtendedRetrieverSpanRecord

                            return ExtendedRetrieverSpanRecord.from_dict(data)
                        except:  # noqa: E722
                            pass
                    elif type_value == "session":
                        try:
                            from ..models.extended_session_record import ExtendedSessionRecord

                            return ExtendedSessionRecord.from_dict(data)
                        except:  # noqa: E722
                            pass

                # Fallback to standard union parsing
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    return ExtendedAgentSpanRecordWithChildren.from_dict(data)

                except:  # noqa: E722
                    pass
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    return ExtendedWorkflowSpanRecordWithChildren.from_dict(data)

                except:  # noqa: E722
                    pass
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    return ExtendedLlmSpanRecord.from_dict(data)

                except:  # noqa: E722
                    pass
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    return ExtendedToolSpanRecordWithChildren.from_dict(data)

                except:  # noqa: E722
                    pass
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    return ExtendedRetrieverSpanRecordWithChildren.from_dict(data)

                except:  # noqa: E722
                    pass
                # If we reach here, none of the parsers succeeded
                discriminator_info = f" (type={data.get('type')})" if isinstance(data, dict) and "type" in data else ""
                raise ValueError(f"Could not parse union type for spans_item{discriminator_info}")

            spans_item = _parse_spans_item(spans_item_data)

            spans.append(spans_item)

        type_ = cast(Union[Literal["retriever"], Unset], d.pop("type", UNSET))
        if type_ != "retriever" and not isinstance(type_, Unset):
            raise ValueError(f"type must match const 'retriever', got '{type_}'")

        input_ = d.pop("input", UNSET)

        def _parse_redacted_input(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        redacted_input = _parse_redacted_input(d.pop("redacted_input", UNSET))

        output = []
        _output = d.pop("output", UNSET)
        for output_item_data in _output or []:
            output_item = Document.from_dict(output_item_data)

            output.append(output_item)

        def _parse_redacted_output(data: object) -> Union[None, Unset, list["Document"]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                redacted_output_type_0 = []
                _redacted_output_type_0 = data
                for redacted_output_type_0_item_data in _redacted_output_type_0:
                    redacted_output_type_0_item = Document.from_dict(redacted_output_type_0_item_data)

                    redacted_output_type_0.append(redacted_output_type_0_item)

                return redacted_output_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list["Document"]], data)

        redacted_output = _parse_redacted_output(d.pop("redacted_output", UNSET))

        name = d.pop("name", UNSET)

        _created_at = d.pop("created_at", UNSET)
        created_at: Union[Unset, datetime.datetime]
        created_at = UNSET if isinstance(_created_at, Unset) else isoparse(_created_at)

        _user_metadata = d.pop("user_metadata", UNSET)
        user_metadata: Union[Unset, ExtendedRetrieverSpanRecordWithChildrenUserMetadata]
        if isinstance(_user_metadata, Unset):
            user_metadata = UNSET
        else:
            user_metadata = ExtendedRetrieverSpanRecordWithChildrenUserMetadata.from_dict(_user_metadata)

        tags = cast(list[str], d.pop("tags", UNSET))

        def _parse_status_code(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        status_code = _parse_status_code(d.pop("status_code", UNSET))

        _metrics = d.pop("metrics", UNSET)
        metrics: Union[Unset, Metrics]
        metrics = UNSET if isinstance(_metrics, Unset) else Metrics.from_dict(_metrics)

        def _parse_external_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        external_id = _parse_external_id(d.pop("external_id", UNSET))

        def _parse_dataset_input(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        dataset_input = _parse_dataset_input(d.pop("dataset_input", UNSET))

        def _parse_dataset_output(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        dataset_output = _parse_dataset_output(d.pop("dataset_output", UNSET))

        _dataset_metadata = d.pop("dataset_metadata", UNSET)
        dataset_metadata: Union[Unset, ExtendedRetrieverSpanRecordWithChildrenDatasetMetadata]
        if isinstance(_dataset_metadata, Unset):
            dataset_metadata = UNSET
        else:
            dataset_metadata = ExtendedRetrieverSpanRecordWithChildrenDatasetMetadata.from_dict(_dataset_metadata)

        def _parse_trace_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        trace_id = _parse_trace_id(d.pop("trace_id", UNSET))

        def _parse_updated_at(data: object) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                return isoparse(data)

            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        updated_at = _parse_updated_at(d.pop("updated_at", UNSET))

        def _parse_has_children(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        has_children = _parse_has_children(d.pop("has_children", UNSET))

        def _parse_metrics_batch_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        metrics_batch_id = _parse_metrics_batch_id(d.pop("metrics_batch_id", UNSET))

        def _parse_session_batch_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        session_batch_id = _parse_session_batch_id(d.pop("session_batch_id", UNSET))

        _feedback_rating_info = d.pop("feedback_rating_info", UNSET)
        feedback_rating_info: Union[Unset, ExtendedRetrieverSpanRecordWithChildrenFeedbackRatingInfo]
        if isinstance(_feedback_rating_info, Unset):
            feedback_rating_info = UNSET
        else:
            feedback_rating_info = ExtendedRetrieverSpanRecordWithChildrenFeedbackRatingInfo.from_dict(
                _feedback_rating_info
            )

        def _parse_metric_info(
            data: object,
        ) -> Union["ExtendedRetrieverSpanRecordWithChildrenMetricInfoType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            # Discriminator-aware parsing for Extended*Record types
            if isinstance(data, dict) and "type" in data:
                type_value = data.get("type")

                # Hardcoded discriminator mapping for Extended*Record types
                if type_value == "trace":
                    try:
                        from ..models.extended_trace_record import ExtendedTraceRecord

                        return ExtendedTraceRecord.from_dict(data)
                    except:  # noqa: E722
                        pass
                elif type_value == "agent":
                    try:
                        from ..models.extended_agent_span_record import ExtendedAgentSpanRecord

                        return ExtendedAgentSpanRecord.from_dict(data)
                    except:  # noqa: E722
                        pass
                elif type_value == "workflow":
                    try:
                        from ..models.extended_workflow_span_record import ExtendedWorkflowSpanRecord

                        return ExtendedWorkflowSpanRecord.from_dict(data)
                    except:  # noqa: E722
                        pass
                elif type_value == "llm":
                    try:
                        from ..models.extended_llm_span_record import ExtendedLlmSpanRecord

                        return ExtendedLlmSpanRecord.from_dict(data)
                    except:  # noqa: E722
                        pass
                elif type_value == "tool":
                    try:
                        from ..models.extended_tool_span_record import ExtendedToolSpanRecord

                        return ExtendedToolSpanRecord.from_dict(data)
                    except:  # noqa: E722
                        pass
                elif type_value == "retriever":
                    try:
                        from ..models.extended_retriever_span_record import ExtendedRetrieverSpanRecord

                        return ExtendedRetrieverSpanRecord.from_dict(data)
                    except:  # noqa: E722
                        pass
                elif type_value == "session":
                    try:
                        from ..models.extended_session_record import ExtendedSessionRecord

                        return ExtendedSessionRecord.from_dict(data)
                    except:  # noqa: E722
                        pass

            # Fallback to standard union parsing
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                return ExtendedRetrieverSpanRecordWithChildrenMetricInfoType0.from_dict(data)

            except:  # noqa: E722
                pass
            return cast(Union["ExtendedRetrieverSpanRecordWithChildrenMetricInfoType0", None, Unset], data)

        metric_info = _parse_metric_info(d.pop("metric_info", UNSET))

        is_complete = d.pop("is_complete", UNSET)

        def _parse_step_number(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        step_number = _parse_step_number(d.pop("step_number", UNSET))

        extended_retriever_span_record_with_children = cls(
            id=id,
            session_id=session_id,
            project_id=project_id,
            run_id=run_id,
            parent_id=parent_id,
            spans=spans,
            type_=type_,
            input_=input_,
            redacted_input=redacted_input,
            output=output,
            redacted_output=redacted_output,
            name=name,
            created_at=created_at,
            user_metadata=user_metadata,
            tags=tags,
            status_code=status_code,
            metrics=metrics,
            external_id=external_id,
            dataset_input=dataset_input,
            dataset_output=dataset_output,
            dataset_metadata=dataset_metadata,
            trace_id=trace_id,
            updated_at=updated_at,
            has_children=has_children,
            metrics_batch_id=metrics_batch_id,
            session_batch_id=session_batch_id,
            feedback_rating_info=feedback_rating_info,
            metric_info=metric_info,
            is_complete=is_complete,
            step_number=step_number,
        )

        extended_retriever_span_record_with_children.additional_properties = d
        return extended_retriever_span_record_with_children

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import logging
import uuid
from datetime import datetime, timezone
from typing import Any, Optional, cast

from agents import Span, Trace, TracingProcessor
from agents.tracing import ResponseSpanData, get_current_span, get_trace_provider

from galileo import GalileoLogger, galileo_context
from galileo.schema.handlers import Node
from galileo.utils import _get_timestamp
from galileo.utils.catch_log import DecorateAllMethods
from galileo.utils.openai_agents import (
    GalileoCustomSpan,
    _extract_llm_data,
    _extract_tool_data,
    _extract_workflow_data,
    _map_span_name,
    _map_span_type,
)
from galileo.utils.serialization import convert_time_delta_to_ns, convert_to_string_dict, serialize_to_str
from galileo_core.schemas.logging.span import LlmMetrics, LlmSpan
from galileo_core.schemas.logging.span import Span as GalileoSpan

_logger = logging.getLogger(__name__)


class GalileoTracingProcessor(TracingProcessor, DecorateAllMethods):
    """
    OpenAI Agents TracingProcessor for logging traces to Galileo.

    Builds a tree of spans during agent execution and logs them hierarchically
    to Galileo upon trace completion.

    Attributes
    ----------
    _galileo_logger : GalileoLogger
        The Galileo logger instance.
    _flush_on_trace_end : bool
        Whether to automatically flush the log batch to Galileo when a trace ends.
    _nodes : dict[str, Node]
        Stores Node objects keyed by their OpenAI span_id or trace_id (for root).
    """

    def __init__(self, galileo_logger: Optional[GalileoLogger] = None, flush_on_trace_end: bool = True):
        """
        OpenAI Agents TracingProcessor for logging traces to Galileo.

        Parameters
        ----------
        galileo_logger : Optional[GalileoLogger]
            The Galileo logger instance. If None, a default instance is created.
        flush_on_trace_end : bool
            Whether to automatically flush the log batch to Galileo when a trace ends.
        """
        self._galileo_logger: GalileoLogger = galileo_logger or galileo_context.get_logger_instance()
        self._flush_on_trace_end: bool = flush_on_trace_end
        self._nodes: dict[str, Node] = {}
        self._last_output: Any = None
        self._last_status_code: Optional[int] = None
        self._first_input: Any = None

    def on_trace_start(self, trace: Trace) -> None:
        """Called when an OpenAI Agent trace starts."""
        node = Node(
            node_type="agent",
            run_id=trace.trace_id,
            span_params={
                "start_time": _get_timestamp(),
                "start_time_iso": datetime.now(timezone.utc).isoformat(),
                "name": trace.name,
                "metadata": convert_to_string_dict(trace.metadata),
            },
        )
        self._nodes[trace.trace_id] = node

    def on_trace_end(self, trace: Trace) -> None:
        """Called when an OpenAI Agent trace ends."""
        node = self._nodes.get(trace.trace_id)
        if not node:
            _logger.warning(f"End called for unknown trace_id {trace.trace_id}")
            return

        node.span_params["duration_ns"] = convert_time_delta_to_ns(_get_timestamp() - node.span_params["start_time"])

        # Log the trace to Galileo
        self._commit_trace(trace)
        self._galileo_logger.conclude(output=self._last_output, status_code=self._last_status_code)
        self._nodes = {}

        self._last_output = None
        self._last_status_code = None
        self._first_input = None

        # Optionally flush the log batch
        if self._flush_on_trace_end:
            self._galileo_logger.flush()

    def _commit_trace(self, trace: Trace) -> None:
        if not self._nodes:
            _logger.warning("No nodes to commit")
            return

        root_node = self._nodes.get(trace.trace_id)
        if root_node:
            self._log_node_tree(root_node, first_node=True)
        else:
            _logger.warning(f"Root node {trace.trace_id} not found")
        self._galileo_logger.conclude(output=self._last_output, status_code=self._last_status_code)

    def _log_node_tree(self, node: Node, first_node: bool = False) -> None:
        """
        Log a node and its children recursively.

        Parameters
        ----------
        node : Node
            The node to log.
        first_node : bool
            Whether this is the root trace node.
        """
        is_workflow_span = False
        input = node.span_params.get("input", "")
        output = node.span_params.get("output", "")
        name = node.span_params.get("name")
        metadata = node.span_params.get("metadata")
        tags = node.span_params.get("tags")
        start_time_iso = datetime.fromisoformat(node.span_params.get("start_time_iso", ""))

        # Convert metadata to a dict[str, str]
        if metadata is not None:
            metadata = convert_to_string_dict(metadata)
        if first_node:
            self._galileo_logger.add_trace(
                input=self._first_input or "Agent Workflow",
                output=self._last_output,
                duration_ns=node.span_params.get("duration_ns"),
                created_at=start_time_iso,
                name="Trace",
                tags=tags,
            )
        # Log the current node based on its type
        elif node.node_type in ("agent", "chain", "workflow"):
            self._galileo_logger.add_workflow_span(
                input=input or node.node_type.capitalize() + " Step",
                output=output,
                name=name,
                metadata=metadata,
                tags=tags,
                created_at=start_time_iso,
                duration_ns=node.span_params.get("duration_ns"),
            )
            is_workflow_span = True
        elif node.node_type in ("llm", "chat"):
            tools = node.span_params.get("tools")

            span = LlmSpan(
                input=input,
                output=output,
                name=name or "LLM Span",
                created_at=start_time_iso,
                user_metadata=metadata or {},
                tags=tags or [],
                metrics=LlmMetrics.model_validate(
                    {
                        "duration_ns": node.span_params.get("duration_ns"),
                        "num_input_tokens": node.span_params.get("num_input_tokens"),
                        "num_output_tokens": node.span_params.get("num_output_tokens"),
                        "num_total_tokens": node.span_params.get("num_total_tokens"),
                        "time_to_first_token_ns": node.span_params.get("time_to_first_token_ns"),
                        "num_reasoning_tokens": node.span_params.get("num_reasoning_tokens"),
                        "num_cached_input_tokens": node.span_params.get("num_cached_input_tokens"),
                    }
                ),
                tools=tools,
                model=node.span_params.get(
                    "model", node.span_params.get("metadata", {}).get("response_metadata", {}).get("model")
                ),
                temperature=node.span_params.get(
                    "temperature", node.span_params.get("metadata", {}).get("response_metadata", {}).get("temperature")
                ),
                status_code=node.span_params.get("status_code", 200),
                id=uuid.uuid4(),
            )
            self._galileo_logger.add_child_span_to_parent(span)
        elif node.node_type == "retriever":
            self._galileo_logger.add_retriever_span(
                input=input,
                output=output,
                name=name,
                metadata=metadata,
                tags=tags,
                created_at=start_time_iso,
                duration_ns=node.span_params.get("duration_ns"),
            )
        elif node.node_type == "tool":
            self._galileo_logger.add_tool_span(
                input=input or node.node_type,
                output=output,
                name=name,
                metadata=metadata,
                tags=tags,
                created_at=start_time_iso,
                duration_ns=node.span_params.get("duration_ns"),
            )
        else:
            _logger.warning(f"Unknown node type: {node.node_type}")

        # Process all child nodes
        last_child = None
        for child_id in node.children:
            child_node = self._nodes.get(child_id)
            if child_node:
                self._log_node_tree(child_node)
                last_child = child_node
            else:
                _logger.warning(f"Child node {child_id} not found")

        # Conclude workflow span. Use the last child's output if necessary
        if is_workflow_span:
            output = output or (last_child.span_params.get("output", "") if last_child else "")
            error = node.span_params.get("error")
            status_code = node.span_params.get("status_code", 200)
            if error:
                output = error
                status_code = 500
            self._galileo_logger.conclude(output=serialize_to_str(output), status_code=status_code)
            self._last_status_code = status_code
            self._last_output = output

    def on_span_start(self, span: Span[Any]) -> None:
        """Called when an OpenAI Agent span starts."""
        span_id = span.span_id
        trace_id = span.trace_id
        parent_id = span.parent_id or span.trace_id  # Parent is previous span or root trace

        if span_id in self._nodes:
            _logger.warning(f"Span node already exists for span_id {span_id}, overwriting...")

        # Determine span type and name
        galileo_type = _map_span_type(span.span_data)
        span_name = _map_span_name(span)

        # Extract initial data based on type
        initial_params: dict[str, Any] = {
            "name": span_name,
            "start_time_iso": span.started_at or datetime.now(timezone.utc).isoformat(),
        }
        if galileo_type in ["llm", "chat"]:
            llm_data = _extract_llm_data(span.span_data)
            initial_params.update(
                {
                    "input": llm_data.get("input"),
                    "model": llm_data.get("model"),
                    "temperature": llm_data.get("temperature"),
                    "tools": llm_data.get("tools"),
                    "model_parameters": llm_data.get("model_parameters"),
                    "metadata": llm_data.get("metadata", {}),
                    "status_code": llm_data.get("status_code", 200),
                }
            )
            if not self._first_input and initial_params.get("input") != serialize_to_str(None):
                self._first_input = initial_params.get("input")
        elif galileo_type == "tool":
            tool_data = _extract_tool_data(span.span_data)
            initial_params.update(
                {
                    "input": tool_data.get("input"),
                    "metadata": tool_data.get("metadata", {}),
                    "status_code": tool_data.get("status_code", 200),
                }
            )
        elif galileo_type == "workflow":
            wf_data = _extract_workflow_data(span.span_data)
            initial_params.update(
                {
                    "input": wf_data.get("input"),
                    "metadata": wf_data.get("metadata", {}),
                    "status_code": wf_data.get("status_code", 200),
                }
            )
        elif galileo_type == "galileo_custom":
            custom_span = cast(GalileoCustomSpan, span.span_data)
            initial_params.update(
                {
                    "input": custom_span.span.input,
                    "output": custom_span.span.output,
                    "metadata": custom_span.span.user_metadata or {},
                    "tags": custom_span.span.tags,
                    "status_code": custom_span.span.status_code,
                }
            )
            galileo_type = custom_span.span.type.value

        if galileo_type == "galileo_custom":
            galileo_type = "workflow"

        # Create the node
        node = Node(node_type=galileo_type, span_params=initial_params, run_id=span_id, parent_run_id=parent_id)
        self._nodes[span_id] = node

        # Add to parent's children list
        parent_node = self._nodes.get(parent_id)
        if not parent_node:
            _logger.warning(f"Parent node {parent_id} not found for span {span_id} in trace {trace_id}")
            return
        parent_node.children.append(span_id)

    def on_span_end(self, span: Span[Any]) -> None:
        """Called when an OpenAI Agent span ends."""
        span_id = span.span_id
        node = self._nodes.get(span_id)
        if not node:
            _logger.warning(f"End called for unknown span_id {span_id}")
            return

        node.span_params["name"] = _map_span_name(span)

        # Update node with final data
        galileo_type = node.node_type
        end_params: dict[str, Any] = {"end_time_iso": span.ended_at or datetime.now(timezone.utc).isoformat()}
        end_params["duration_ns"] = convert_time_delta_to_ns(
            datetime.fromisoformat(span.ended_at) - datetime.fromisoformat(node.span_params["start_time_iso"])
        )

        if galileo_type == "llm":
            llm_data = _extract_llm_data(span.span_data)
            end_params.update(
                {
                    **llm_data,
                    "metadata": {**node.span_params.get("metadata", {}), **llm_data.get("metadata", {})},
                    "status_code": llm_data.get("status_code", node.span_params.get("status_code", 200)),
                }
            )
            # Ensure input is preserved if it wasn't available at start
            if node.span_params.get("input") is None:
                node.span_params["input"] = llm_data.get("input")
                if (
                    not self._first_input
                    and node.span_params["input"]
                    and node.span_params["input"] != serialize_to_str(None)
                ):
                    self._first_input = node.span_params["input"]

            # Extract embedded tool calls and merge with existing tool definitions
            if isinstance(span.span_data, ResponseSpanData) and span.span_data.response:
                embedded_tool_calls = self._extract_embedded_tool_calls(span.span_data.response)
                if embedded_tool_calls:
                    existing_tools = llm_data.get("tools") or []
                    if not isinstance(existing_tools, list):
                        existing_tools = []
                    end_params["tools"] = existing_tools + embedded_tool_calls

        elif galileo_type == "tool":
            tool_data = _extract_tool_data(span.span_data)
            end_params.update(
                {
                    **tool_data,
                    "metadata": {**node.span_params.get("metadata", {}), **tool_data.get("metadata", {})},
                    "status_code": tool_data.get("status_code", node.span_params.get("status_code", 200)),
                }
            )
            if node.span_params.get("input") is None:
                node.span_params["input"] = tool_data.get("input")

        elif galileo_type == "workflow":
            wf_data = _extract_workflow_data(span.span_data)
            end_params.update(
                {
                    **wf_data,
                    "metadata": {**node.span_params.get("metadata", {}), **wf_data.get("metadata", {})},
                    "status_code": wf_data.get("status_code", node.span_params.get("status_code", 200)),
                }
            )
            # Workflow output might only be known at the end
            if node.span_params.get("output") is None:
                node.span_params["output"] = wf_data.get("output")

        # Handle errors
        if error := span.error:
            end_params["error"] = error  # Store raw error
            end_params["status_code"] = 500  # Indicate error status
            end_params["metadata"] = {
                **end_params.get("metadata", {}),
                "error_message": error.get("message", str(error)),
                "error_type": error.get("type", type(error).__name__),
                "error_details": serialize_to_str(error.get("data")),
            }

        # Update the node's parameters
        node.span_params.update(end_params)

    def shutdown(self) -> None:
        """Called when the application stops. Flushes any remaining logs."""
        self._galileo_logger.flush()

    def force_flush(self) -> None:
        """Forces an immediate flush of all queued traces/spans."""
        self._galileo_logger.flush()

    def _extract_embedded_tool_calls(self, response: Any) -> list[dict[str, Any]]:
        """Extract embedded tool calls from response.output."""
        if not response or not hasattr(response, "output"):
            return []

        output = response.output
        if not isinstance(output, list):
            return []

        tool_calls = []
        for item in output:
            if hasattr(item, "model_dump"):
                try:
                    item_dict = item.model_dump()
                except Exception:  # noqa: S112
                    continue
            elif isinstance(item, dict):
                item_dict = item
            else:
                continue

            item_type = item_dict.get("type", "")
            if item_type not in (
                "code_interpreter_call",
                "file_search_call",
                "web_search_call",
                "computer_call",
                "custom_tool_call",
            ):
                continue

            tool_call = {
                "type": "function",
                "function": {"name": self._get_tool_name_from_type(item_type)},
                "tool_call_id": item_dict.get("id") or item_dict.get("call_id"),
                "tool_call_type": item_type,
                "tool_call_input": self._extract_tool_input(item_dict, item_type),
                "tool_call_output": self._extract_tool_output(item_dict, item_type),
                "tool_call_status": item_dict.get("status", "completed"),
            }
            tool_calls.append(tool_call)

        return tool_calls

    def _get_tool_name_from_type(self, item_type: str) -> str:
        """Map OpenAI tool call type to tool name."""
        type_to_name = {
            "code_interpreter_call": "code_interpreter",
            "file_search_call": "file_search",
            "web_search_call": "web_search",
            "computer_call": "computer",
            "custom_tool_call": "custom_tool",
        }
        return type_to_name.get(item_type, item_type)

    def _extract_tool_input(self, item_dict: dict[str, Any], item_type: str) -> str:
        """Extract input from tool call item per OpenAI schema."""
        if item_type == "code_interpreter_call":
            return serialize_to_str(item_dict.get("code"))
        if item_type == "file_search_call":
            queries = item_dict.get("queries")
            return serialize_to_str(queries)
        if item_type == "web_search_call":
            action = item_dict.get("action", {})
            if isinstance(action, dict) and action.get("type") == "search":
                return serialize_to_str(action.get("query"))
            return serialize_to_str(action)
        if item_type == "custom_tool_call":
            return serialize_to_str(item_dict.get("input"))
        return serialize_to_str(item_dict.get("input") or item_dict.get("action"))

    def _extract_tool_output(self, item_dict: dict[str, Any], item_type: str) -> str:
        """Extract output from tool call item per OpenAI schema."""
        if item_type == "code_interpreter_call":
            outputs = item_dict.get("outputs")
            if outputs and isinstance(outputs, list):
                output_parts = []
                for output_item in outputs:
                    if isinstance(output_item, dict):
                        if output_item.get("type") == "logs":
                            output_parts.append(output_item.get("logs", ""))
                        elif output_item.get("type") == "image":
                            output_parts.append(output_item.get("url", ""))
                    else:
                        output_parts.append(str(output_item))
                return serialize_to_str("\n".join(output_parts) if output_parts else None)
            return serialize_to_str(None)
        if item_type == "file_search_call":
            return serialize_to_str(item_dict.get("results"))
        if item_type == "web_search_call":
            return serialize_to_str(item_dict.get("action"))
        if item_type == "custom_tool_call":
            return serialize_to_str(None)
        return serialize_to_str(item_dict.get("output") or item_dict.get("results"))

    @staticmethod
    def add_galileo_custom_span(span: GalileoSpan) -> Span[GalileoCustomSpan]:
        """Add a Galileo custom span to the trace."""
        trace_provider = get_trace_provider()
        current_span = get_current_span()
        custom_span = GalileoCustomSpan(span, span.user_metadata)
        return trace_provider.create_span(custom_span, parent=current_span)



import datetime
from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

T = TypeVar("T", bound="RunTagDB")


@_attrs_define
class RunTagDB:
    """
    Attributes
    ----------
        key (str):
        value (str):
        tag_type (str):
        project_id (str):
        run_id (str):
        created_by (str):
        id (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
    """

    key: str
    value: str
    tag_type: str
    project_id: str
    run_id: str
    created_by: str
    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        key = self.key

        value = self.value

        tag_type = self.tag_type

        project_id = self.project_id

        run_id = self.run_id

        created_by = self.created_by

        id = self.id

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "key": key,
                "value": value,
                "tag_type": tag_type,
                "project_id": project_id,
                "run_id": run_id,
                "created_by": created_by,
                "id": id,
                "created_at": created_at,
                "updated_at": updated_at,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        key = d.pop("key")

        value = d.pop("value")

        tag_type = d.pop("tag_type")

        project_id = d.pop("project_id")

        run_id = d.pop("run_id")

        created_by = d.pop("created_by")

        id = d.pop("id")

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        run_tag_db = cls(
            key=key,
            value=value,
            tag_type=tag_type,
            project_id=project_id,
            run_id=run_id,
            created_by=created_by,
            id=id,
            created_at=created_at,
            updated_at=updated_at,
        )

        run_tag_db.additional_properties = d
        return run_tag_db

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any

from galileo.__future__.shared.base import StateManagementMixin, SyncState
from galileo.__future__.shared.exceptions import APIError, ValidationError
from galileo.projects import Projects

if TYPE_CHECKING:
    from galileo.__future__.dataset import Dataset
    from galileo.__future__.experiment import Experiment
    from galileo.__future__.log_stream import LogStream
    from galileo.__future__.prompt import Prompt

logger = logging.getLogger(__name__)


class Project(StateManagementMixin):
    """
    Object-centric interface for Galileo projects.

    This class provides an intuitive way to work with Galileo projects,
    encapsulating project management operations and providing seamless
    integration with log stream management.

    Attributes
    ----------
        created_at (datetime.datetime): When the project was created.
        created_by (str): The user who created the project.
        id (str): The unique project identifier.
        updated_at (datetime.datetime): When the project was last updated.
        bookmark (Union[Unset, bool]): Whether the project is bookmarked.
        name (Union[None, Unset, str]): The project name.
        permissions (Union[Unset, list]): Project permissions.
        type (Union[None, ProjectType, Unset]): The project type.

    Examples
    --------
        # Create a new project locally, then persist
        project = Project(name="My AI Project").create()

        # Get an existing project
        project = Project.get(name="My AI Project")

        # List all projects
        projects = Project.list()

        # Create a log stream for the project
        log_stream = project.create_log_stream(name="Production Logs")

        # List log streams for the project
        log_streams = project.list_log_streams()

        # Access related resources via properties
        for log_stream in project.logstreams:
            print(log_stream.name)

        for experiment in project.experiments:
            print(experiment.name)

        for dataset in project.datasets:
            print(dataset.name)

        for prompt in project.prompts:
            print(prompt.name)

        # Or use the explicit list methods
        datasets = project.list_datasets()
        prompts = project.list_prompts()

        # Delete a project (WARNING: cannot be undone!)
        old_project = Project.get(name="Old Project")
        old_project.delete()
    """

    # Type annotations for instance attributes
    id: str | None
    name: str | None
    created_at: datetime | None
    created_by: str | None
    updated_at: datetime | None
    bookmark: bool | None
    permissions: list[Any] | None
    type: Any | None

    def __str__(self) -> str:
        """String representation of the project."""
        return f"Project(name='{self.name}', id='{self.id}')"

    def __repr__(self) -> str:
        """Detailed string representation of the project."""
        return f"Project(name='{self.name}', id='{self.id}', created_at='{self.created_at}')"

    def __init__(self, name: str | None = None) -> None:
        """
        Initialize a Project instance locally.

        Creates a local project object that exists only in memory until .create()
        is called to persist it to the API.

        Args:
            name (Optional[str]): The name of the project to create.

        Raises
        ------
            ValidationError: If name is not provided.
        """
        super().__init__()
        if name is None:
            raise ValidationError(
                "'name' must be provided to create a project. Use Project.get() to retrieve an existing project."
            )

        # Initialize attributes locally
        self.name = name
        self.id = None
        self.created_at = None
        self.created_by = None
        self.updated_at = None
        self.bookmark = None
        self.permissions = None
        self.type = None

        # Set initial state
        self._set_state(SyncState.LOCAL_ONLY)

    @classmethod
    def _create_empty(cls) -> Project:
        """Internal constructor bypassing __init__ for API hydration."""
        instance = cls.__new__(cls)
        super(Project, instance).__init__()
        return instance

    @classmethod
    def _from_api_response(cls, retrieved_project: Any) -> Project:
        """
        Factory method to create a Project instance from an API response.

        Args:
            retrieved_project: The project data retrieved from the API.

        Returns
        -------
            Project: A new Project instance populated with the API data.
        """
        instance = cls._create_empty()
        instance.created_at = retrieved_project.created_at
        instance.created_by = retrieved_project.created_by
        instance.id = retrieved_project.id
        instance.updated_at = retrieved_project.updated_at
        instance.bookmark = retrieved_project.bookmark
        instance.name = retrieved_project.name
        instance.permissions = retrieved_project.permissions
        instance.type = retrieved_project.type
        # Set state to synced since we just retrieved from API
        instance._set_state(SyncState.SYNCED)
        return instance

    def create(self) -> Project:
        """
        Persist this project to the API.

        Returns
        -------
            Project: This project instance with updated attributes from the API.

        Raises
        ------
            Exception: If the API call fails.

        Examples
        --------
            project = Project(name="My AI Project").create()
            assert project.is_synced()
        """
        if self.name is None:
            raise ValueError("Project name is not set. Cannot create project without a name.")
        try:
            logger.info(f"Project.create: name='{self.name}' - started")
            projects_service = Projects()
            created_project = projects_service.create(name=self.name)

            # Update attributes from response
            self.created_at = created_project.created_at
            self.created_by = created_project.created_by
            self.id = created_project.id
            self.updated_at = created_project.updated_at
            self.bookmark = created_project.bookmark
            self.name = created_project.name
            self.permissions = created_project.permissions
            self.type = created_project.type

            # Set state to synced
            self._set_state(SyncState.SYNCED)
            logger.info(f"Project.create: id='{self.id}' - completed")
            return self
        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"Project.create: name='{self.name}' - failed: {e}")
            raise

    @classmethod
    def get(cls, *, id: str | None = None, name: str | None = None) -> Project | None:
        """
        Get an existing project by ID or name.

        Args:
            id (Optional[str]): The project ID.
            name (Optional[str]): The project name.

        Returns
        -------
            Optional[Project]: The project if found, None otherwise.

        Raises
        ------
            ValidationError: If neither or both id and name are provided.

        Examples
        --------
            # Get by name
            project = Project.get(name="My AI Project")

            # Get by ID
            project = Project.get(id="project-123")
        """
        try:
            projects_service = Projects()
            retrieved_project = projects_service.get(id=id, name=name)
            if retrieved_project is None:
                return None

            return cls._from_api_response(retrieved_project)
        except Exception as e:
            logger.error("Project.get: id='%s' name='%s' - failed: %s", id, name, str(e))
            raise APIError("Failed to retrieve project: %s", original_error=e) from e

    @classmethod
    def list(cls) -> list[Project]:
        """
        List all available projects.

        Returns
        -------
            List[Project]: A list of all projects.

        Examples
        --------
            projects = Project.list()
            for project in projects:
                # Process each project
                pass
        """
        projects_service = Projects()
        retrieved_projects = projects_service.list()

        return [cls._from_api_response(retrieved_project) for retrieved_project in retrieved_projects]

    def create_log_stream(self, name: str) -> LogStream:
        """
        Create a new log stream for this project.

        Args:
            name (str): The name of the log stream to create.

        Returns
        -------
            LogStream: The created log stream.

        Examples
        --------
            project = Project.get(name="My AI Project")
            log_stream = project.create_log_stream(name="Production Logs")
        """
        if self.id is None:
            raise ValueError("Project ID is not set. Cannot create log stream for a local-only project.")

        # Use the LogStream pattern to avoid duplication
        return LogStream(name=name, project_id=self.id).create()

    def list_log_streams(self) -> list[LogStream]:  # type: ignore[valid-type]
        """
        List all log streams for this project.

        Returns
        -------
            List[LogStream]: A list of log streams belonging to this project.

        Examples
        --------
            project = Project.get(name="My AI Project")
            log_streams = project.list_log_streams()
            for stream in log_streams:
                # Process each log stream
                pass
        """
        if self.id is None:
            raise ValueError("Project ID is not set. Cannot list log streams for a local-only project.")

        # Use the LogStream pattern to avoid duplication
        return LogStream.list(project_id=self.id)

    def list_experiments(self) -> list[Experiment]:  # type: ignore[valid-type]
        """
        List all experiments for this project.

        Returns
        -------
            List[Experiment]: A list of experiments belonging to this project.

        Examples
        --------
            project = Project.get(name="My AI Project")
            experiments = project.list_experiments()
            for exp in experiments:
                # Process each experiment
                pass
        """
        if self.id is None:
            raise ValueError("Project ID is not set. Cannot list experiments for a local-only project.")

        # Use the Experiment pattern to avoid duplication
        return Experiment.list(project_id=self.id)

    def list_datasets(self) -> list[Dataset]:  # type: ignore[valid-type]
        """
        List all datasets used in this project.

        Returns
        -------
            List[Dataset]: A list of datasets used in this project.

        Examples
        --------
            project = Project.get(name="My AI Project")
            datasets = project.list_datasets()
            for dataset in datasets:
                # Process each dataset
                pass
        """
        if self.id is None:
            raise ValueError("Project ID is not set. Cannot list datasets for a local-only project.")

        # Use the Dataset pattern to avoid duplication
        return Dataset.list(project_id=self.id)

    def list_prompts(self) -> list[Prompt]:  # type: ignore[valid-type]
        """
        List all prompts used in this project.

        Returns
        -------
            List[Prompt]: A list of prompts used in this project.

        Examples
        --------
            project = Project.get(name="My AI Project")
            prompts = project.list_prompts()
            for prompt in prompts:
                # Process each prompt
                pass
        """
        if self.id is None:
            raise ValueError("Project ID is not set. Cannot list prompts for a local-only project.")

        # Use the Prompt pattern to avoid duplication
        return Prompt.list(project_id=self.id)

    @property
    def logstreams(self) -> list[LogStream]:  # type: ignore[valid-type]
        """
        Property to access log streams for this project.

        This is a read-only property that returns the current list of log streams.
        To create new log streams, use create_log_stream().

        Returns
        -------
            List[LogStream]: A list of log streams belonging to this project.

        Examples
        --------
            project = Project.get(name="My AI Project")
            for stream in project.logstreams:
                print(stream.name)
        """
        return self.list_log_streams()

    @property
    def experiments(self) -> list[Experiment]:  # type: ignore[valid-type]
        """
        Property to access experiments for this project.

        This is a read-only property that returns the current list of experiments.

        Returns
        -------
            List[Experiment]: A list of experiments belonging to this project.

        Examples
        --------
            project = Project.get(name="My AI Project")
            for exp in project.experiments:
                print(exp.name)
        """
        return self.list_experiments()

    @property
    def datasets(self) -> list[Dataset]:  # type: ignore[valid-type]
        """
        Property to access datasets used in this project.

        This is a read-only property that returns the datasets associated with this project.

        Returns
        -------
            List[Dataset]: A list of datasets used in this project.

        Examples
        --------
            project = Project.get(name="My AI Project")
            for dataset in project.datasets:
                print(dataset.name)
        """
        return self.list_datasets()

    @property
    def prompts(self) -> list[Prompt]:  # type: ignore[valid-type]
        """
        Property to access prompts used in this project.

        This is a read-only property that returns the prompts associated with this project.

        Returns
        -------
            List[Prompt]: A list of prompts used in this project.

        Examples
        --------
            project = Project.get(name="My AI Project")
            for prompt in project.prompts:
                print(prompt.name)
        """
        return self.list_prompts()

    def refresh(self) -> None:
        """
        Refresh this project's state from the API.

        Updates all attributes with the latest values from the remote API
        and sets the state to SYNCED.

        Raises
        ------
            Exception: If the API call fails or the project no longer exists.

        Examples
        --------
            project.refresh()
            assert project.is_synced()
        """
        if self.id is None:
            raise ValueError("Project ID is not set. Cannot refresh a local-only project.")
        try:
            logger.debug(f"Project.refresh: id='{self.id}' - started")
            projects_service = Projects()
            retrieved_project = projects_service.get(id=self.id)

            if retrieved_project is None:
                raise ValueError(f"Project with id '{self.id}' no longer exists")

            # Update all attributes from response
            self.created_at = retrieved_project.created_at
            self.created_by = retrieved_project.created_by
            self.id = retrieved_project.id
            self.updated_at = retrieved_project.updated_at
            self.bookmark = retrieved_project.bookmark
            self.name = retrieved_project.name
            self.permissions = retrieved_project.permissions
            self.type = retrieved_project.type

            # Set state to synced
            self._set_state(SyncState.SYNCED)
            logger.debug(f"Project.refresh: id='{self.id}' - completed")
        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"Project.refresh: id='{self.id}' - failed: {e}")
            raise

    def delete(self) -> None:
        """
        Delete this project.

        This is a destructive operation that permanently removes the project
        and all associated data (experiments, log streams, datasets, traces, etc.)
        from the API.

        WARNING: This operation cannot be undone!

        After successful deletion, the object state is set to DELETED. The local
        object still exists in memory but no longer represents a remote resource.

        Raises
        ------
            ValueError: If the project ID is not set.
            Exception: If the API call fails.

        Examples
        --------
            # Delete a project
            project = Project.get(name="Old Project")
            project.delete()
            assert project.is_deleted()

            # After deletion, the project no longer exists remotely
            # The local object is marked as DELETED
            print(project.sync_state)  # SyncState.DELETED
        """
        if self.id is None:
            raise ValueError("Project ID is not set. Cannot delete a local-only project.")

        try:
            logger.info(f"Project.delete: name='{self.name}' id='{self.id}' - started")
            projects_service = Projects()
            projects_service.delete_project(id=self.id)
            # Set state to deleted after successful deletion
            self._set_state(SyncState.DELETED)
            logger.info(f"Project.delete: name='{self.name}' id='{self.id}' - completed")
        except Exception as e:
            self._set_state(SyncState.FAILED_SYNC, error=e)
            logger.error(f"Project.delete: name='{self.name}' id='{self.id}' - failed: {e}")
            raise

    def save(self) -> Project:
        """
        Save changes to this project.

        This method is a placeholder for future functionality to update
        project properties.

        Returns
        -------
            Project: This project instance.

        Raises
        ------
            NotImplementedError: This functionality is not yet implemented.
        """
        raise NotImplementedError(
            "Project updates are not yet implemented. Use specific methods to modify project state."
        )


# Import at end to avoid circular import (log_stream.py imports Project)
from galileo.__future__.dataset import Dataset  # noqa: E402
from galileo.__future__.experiment import Experiment  # noqa: E402
from galileo.__future__.log_stream import LogStream  # noqa: E402
from galileo.__future__.prompt import Prompt  # noqa: E402



"""
Galileo Future API.

This package provides the next-generation object-centric API for Galileo.
"""

from galileo.__future__.configuration import Configuration
from galileo.__future__.dataset import Dataset
from galileo.__future__.experiment import Experiment
from galileo.__future__.integration import Integration
from galileo.__future__.log_stream import LogStream
from galileo.__future__.metric import CodeMetric, GalileoMetric, LlmMetric, LocalMetric, Metric
from galileo.__future__.model import Model
from galileo.__future__.project import Project
from galileo.__future__.prompt import Prompt
from galileo.__future__.shared.exceptions import (
    APIError,
    ConfigurationError,
    GalileoFutureError,
    ResourceConflictError,
    ResourceNotFoundError,
    ValidationError,
)
from galileo.schema.message import Message
from galileo.search import RecordType
from galileo.utils.logging import enable_console_logging
from galileo_core.schemas.logging.llm import MessageRole
from galileo_core.schemas.logging.step import StepType

__all__ = [
    "APIError",
    "CodeMetric",
    "Configuration",
    "ConfigurationError",
    "Dataset",
    "Experiment",
    "GalileoFutureError",
    "GalileoMetric",
    "Integration",
    "LlmMetric",
    "LocalMetric",
    "LogStream",
    "Message",
    "MessageRole",
    "Metric",
    "Model",
    "Project",
    "Prompt",
    "RecordType",
    "ResourceConflictError",
    "ResourceNotFoundError",
    "StepType",
    "ValidationError",
    "enable_console_logging",
]



"""Galileo."""

# flake8: noqa: F401
# ruff: noqa: F401

from galileo.decorator import GalileoDecorator, galileo_context, log, start_session
from galileo.exceptions import GalileoLoggerException
from galileo.logger import GalileoLogger
from galileo.protect import ainvoke_protect, invoke_protect
from galileo.schema.message import Message
from galileo.schema.metrics import GalileoMetrics, GalileoScorers
from galileo.stages import (
    create_protect_stage,
    get_protect_stage,
    pause_protect_stage,
    resume_protect_stage,
    update_protect_stage,
)
from galileo.tracing import get_tracing_headers
from galileo.utils.logging import enable_console_logging
from galileo_core.helpers.api_key import create_api_key, delete_api_key, list_api_keys
from galileo_core.helpers.dependencies import is_dependency_available
from galileo_core.schemas.logging.llm import MessageRole, ToolCall, ToolCallFunction
from galileo_core.schemas.logging.session import Session
from galileo_core.schemas.logging.span import (
    AgentSpan,
    LlmSpan,
    RetrieverSpan,
    Span,
    StepWithChildSpans,
    ToolSpan,
    WorkflowSpan,
)
from galileo_core.schemas.logging.step import StepType
from galileo_core.schemas.logging.trace import Trace
from galileo_core.schemas.protect.execution_status import ExecutionStatus
from galileo_core.schemas.protect.payload import Payload
from galileo_core.schemas.protect.request import Request
from galileo_core.schemas.protect.response import Response
from galileo_core.schemas.protect.ruleset import Ruleset
from galileo_core.schemas.protect.stage import StageType

__version__ = "1.42.0"



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from galileo.utils.headers_data import get_sdk_header
from galileo_core.constants.request_method import RequestMethod
from galileo_core.helpers.api_client import ApiClient

from ... import errors
from ...models.http_validation_error import HTTPValidationError
from ...models.project_db import ProjectDB
from ...models.project_type import ProjectType
from ...types import UNSET, Response, Unset


def _get_kwargs(
    *, project_name: Union[None, Unset, str] = UNSET, type_: Union[None, ProjectType, Unset] = UNSET
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    params: dict[str, Any] = {}

    json_project_name: Union[None, Unset, str]
    json_project_name = UNSET if isinstance(project_name, Unset) else project_name
    params["project_name"] = json_project_name

    json_type_: Union[None, Unset, str]
    if isinstance(type_, Unset):
        json_type_ = UNSET
    elif isinstance(type_, ProjectType):
        json_type_ = type_.value
    else:
        json_type_ = type_
    params["type"] = json_type_

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": RequestMethod.GET,
        "return_raw_response": True,
        "path": "/projects",
        "params": params,
    }

    headers["X-Galileo-SDK"] = get_sdk_header()

    _kwargs["content_headers"] = headers
    return _kwargs


def _parse_response(
    *, client: ApiClient, response: httpx.Response
) -> Optional[Union[HTTPValidationError, list["ProjectDB"]]]:
    if response.status_code == 200:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = ProjectDB.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200

    if response.status_code == 422:
        return HTTPValidationError.from_dict(response.json())

    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    return None


def _build_response(
    *, client: ApiClient, response: httpx.Response
) -> Response[Union[HTTPValidationError, list["ProjectDB"]]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *, client: ApiClient, project_name: Union[None, Unset, str] = UNSET, type_: Union[None, ProjectType, Unset] = UNSET
) -> Response[Union[HTTPValidationError, list["ProjectDB"]]]:
    """Get Projects.

     Gets projects for a user.

    If provided, filters on project_name and project_type.

    DEPRECATED in favor of `get_projects_paginated`.

    Args:
        project_name (Union[None, Unset, str]):
        type_ (Union[None, ProjectType, Unset]):

    Raises
    ------
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns
    -------
        Response[Union[HTTPValidationError, list['ProjectDB']]]
    """
    kwargs = _get_kwargs(project_name=project_name, type_=type_)

    response = client.request(**kwargs)

    return _build_response(client=client, response=response)


def sync(
    *, client: ApiClient, project_name: Union[None, Unset, str] = UNSET, type_: Union[None, ProjectType, Unset] = UNSET
) -> Optional[Union[HTTPValidationError, list["ProjectDB"]]]:
    """Get Projects.

     Gets projects for a user.

    If provided, filters on project_name and project_type.

    DEPRECATED in favor of `get_projects_paginated`.

    Args:
        project_name (Union[None, Unset, str]):
        type_ (Union[None, ProjectType, Unset]):

    Raises
    ------
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns
    -------
        Union[HTTPValidationError, list['ProjectDB']]
    """
    return sync_detailed(client=client, project_name=project_name, type_=type_).parsed


async def asyncio_detailed(
    *, client: ApiClient, project_name: Union[None, Unset, str] = UNSET, type_: Union[None, ProjectType, Unset] = UNSET
) -> Response[Union[HTTPValidationError, list["ProjectDB"]]]:
    """Get Projects.

     Gets projects for a user.

    If provided, filters on project_name and project_type.

    DEPRECATED in favor of `get_projects_paginated`.

    Args:
        project_name (Union[None, Unset, str]):
        type_ (Union[None, ProjectType, Unset]):

    Raises
    ------
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns
    -------
        Response[Union[HTTPValidationError, list['ProjectDB']]]
    """
    kwargs = _get_kwargs(project_name=project_name, type_=type_)

    response = await client.arequest(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *, client: ApiClient, project_name: Union[None, Unset, str] = UNSET, type_: Union[None, ProjectType, Unset] = UNSET
) -> Optional[Union[HTTPValidationError, list["ProjectDB"]]]:
    """Get Projects.

     Gets projects for a user.

    If provided, filters on project_name and project_type.

    DEPRECATED in favor of `get_projects_paginated`.

    Args:
        project_name (Union[None, Unset, str]):
        type_ (Union[None, ProjectType, Unset]):

    Raises
    ------
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns
    -------
        Union[HTTPValidationError, list['ProjectDB']]
    """
    return (await asyncio_detailed(client=client, project_name=project_name, type_=type_)).parsed



from enum import Enum


class DataType(str, Enum):
    BOOLEAN = "boolean"
    DATASET = "dataset"
    FLOATING_POINT = "floating_point"
    INTEGER = "integer"
    PLAYGROUND = "playground"
    PROMPT = "prompt"
    RANK = "rank"
    SCORE_RATING_AGGREGATE = "score_rating_aggregate"
    STAR_RATING_AGGREGATE = "star_rating_aggregate"
    STRING_LIST = "string_list"
    TAG = "tag"
    TAGS_RATING_AGGREGATE = "tags_rating_aggregate"
    TEXT = "text"
    TEXT_RATING_AGGREGATE = "text_rating_aggregate"
    THUMB_RATING_AGGREGATE = "thumb_rating_aggregate"
    TIMESTAMP = "timestamp"
    UUID = "uuid"

    def __str__(self) -> str:
        return str(self.value)



"""
Distributed tracing middleware for Starlette-based applications.

This middleware automatically extracts distributed tracing headers from incoming HTTP requests
and makes them available to the Galileo logger within request handlers.

Works with any ASGI framework built on Starlette:
- FastAPI
- Starlette
- Any other Starlette-based framework

Example usage with FastAPI:
    ```python
    from fastapi import FastAPI
    from galileo.middleware import TracingMiddleware, get_request_logger

    app = FastAPI()
    app.add_middleware(TracingMiddleware)

    @app.post("/process")
    async def process_request(data: dict):
        # Logger automatically continues the distributed trace
        logger = get_request_logger()
        logger.add_workflow_span(input=str(data), name="process_workflow")
        # ... process request ...
        logger.conclude(output="done")
        return {"status": "success"}
    ```

Example usage with Starlette:
    ```python
    from starlette.applications import Starlette
    from starlette.routing import Route
    from galileo.middleware import TracingMiddleware, get_request_logger

    async def homepage(request):
        logger = get_request_logger()
        logger.add_workflow_span(input="homepage", name="homepage_handler")
        logger.conclude(output="success")
        return {"status": "ok"}

    app = Starlette(
        routes=[Route("/", homepage)],
        middleware=[TracingMiddleware]
    )
    ```
"""

import logging
from typing import Any, NoReturn

from galileo.constants.tracing import PARENT_ID_HEADER, TRACE_ID_HEADER
from galileo.decorator import _parent_id_context, _trace_id_context
from galileo.logger import GalileoLogger

_logger = logging.getLogger(__name__)

INSTALL_ERR_MSG = (
    "Starlette is not installed. Install optional middleware dependencies with: pip install galileo[middleware]"
)

try:
    from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
    from starlette.requests import Request
    from starlette.responses import Response
    from starlette.types import ASGIApp
except ImportError:
    # Create stub classes if Starlette is not available
    class BaseHTTPMiddleware:  # type: ignore[no-redef]
        def __init__(self, *args: Any, **kwargs: Any) -> NoReturn:
            raise ImportError(INSTALL_ERR_MSG)

    class Request:  # type: ignore[no-redef]
        pass

    class Response:  # type: ignore[no-redef]
        pass

    class RequestResponseEndpoint:  # type: ignore[no-redef]
        pass

    class ASGIApp:  # type: ignore[no-redef]
        pass


class TracingMiddleware(BaseHTTPMiddleware):
    """
    Middleware that extracts distributed tracing headers from incoming requests.

    This middleware looks for the following headers in incoming HTTP requests:
    - X-Galileo-Trace-ID: The root trace ID
    - X-Galileo-Parent-ID: The parent span/trace ID to attach to

    These values are stored in context variables, making them available to request
    handlers via the `get_request_logger()` function.

    The middleware is compatible with FastAPI and any Starlette-based framework.

    Note: Project and log_stream are configured per service via environment variables
    (GALILEO_PROJECT and GALILEO_LOG_STREAM). They are not propagated via headers,
    following standard distributed tracing patterns.
    """

    def __init__(self, app: ASGIApp) -> None:
        """
        Initialize the tracing middleware.

        Parameters
        ----------
        app : ASGIApp
            The ASGI application
        """
        super().__init__(app)

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        """
        Process the request and extract tracing headers.

        Parameters
        ----------
        request : Request
            The incoming HTTP request
        call_next : RequestResponseEndpoint
            The next middleware or route handler

        Returns
        -------
        Response
            The HTTP response
        """
        # Extract tracing headers from request
        trace_id = request.headers.get(TRACE_ID_HEADER)
        parent_id = request.headers.get(PARENT_ID_HEADER)

        # Store in context variables (thread-safe for async)
        trace_id_token = _trace_id_context.set(trace_id)
        parent_id_token = _parent_id_context.set(parent_id)

        try:
            # Process the request
            return await call_next(request)
        finally:
            # Clean up context variables
            _trace_id_context.reset(trace_id_token)
            _parent_id_context.reset(parent_id_token)


def get_request_logger() -> GalileoLogger:
    """
    Get a request-scoped GalileoLogger configured for distributed mode.

    Note: Distributed mode enables distributed tracing across services by propagating
    trace context and sending updates immediately to the backend.

    This function should be called within a request handler after the TracingMiddleware has
    been registered. It creates a new GalileoLogger instance per request that automatically
    continues the distributed trace from the upstream service.

    The logger is configured using trace context extracted by the middleware:
    - X-Galileo-Trace-ID: Root trace ID
    - X-Galileo-Parent-ID: Parent span/trace ID to attach to

    Project and log_stream are configured per service via environment variables
    (GALILEO_PROJECT and GALILEO_LOG_STREAM), not propagated via headers, following
    standard distributed tracing patterns.

    If no tracing headers were present in the request, a regular logger is returned
    (using GALILEO_PROJECT and GALILEO_LOG_STREAM env vars).

    Note: This creates a new logger per request, unlike the decorator's get_logger_instance()
    which uses a singleton pattern.

    Returns
    -------
    GalileoLogger
        A logger instance configured for the current request's trace context

    Examples
    --------
    ```python
    @app.post("/process")
    async def process_request(data: dict):
        logger = get_request_logger()

        # This span will be attached to the distributed trace
        logger.add_workflow_span(input=str(data), name="process_workflow")
        result = await process(data)
        logger.conclude(output=str(result))

        logger.flush()
        logger.terminate()

        return {"result": result}
    ```

    ```python
    @app.post("/retrieve")
    async def retrieve_endpoint(query: str):
        # Get logger with trace context from upstream service
        logger = get_request_logger()

        # If trace context exists, this creates a workflow span
        # Otherwise, it starts a new trace
        if logger.trace_id:
            logger.add_workflow_span(input=query, name="retrieval_service")
        else:
            logger.start_trace(input=query, name="retrieval_service")

        results = retrieve(query, logger)

        logger.conclude(output=str(results))
        logger.flush()
        logger.terminate()

        return {"results": results}
    ```
    """
    # Get trace context from middleware
    trace_id = _trace_id_context.get()
    parent_id = _parent_id_context.get()

    # Create logger with trace context
    # Project and log_stream come from env vars (GALILEO_PROJECT, GALILEO_LOG_STREAM)
    # If parent_id equals trace_id, it means the parent is the root trace itself,
    # not a span. In this case, we should pass None as span_id to avoid
    # GalileoLoggerException when it tries to look up a span with the trace_id.
    return GalileoLogger(mode="distributed", trace_id=trace_id, span_id=parent_id if parent_id != trace_id else None)
